"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.generator.Utils import tabify

class PopulationGenerator(object):
    """
    Base class for population generators in ANNarchy. Inherited by

    * OpenMPGenerator: single-thread and multi-core implementation
    * CUDAGenerator: gpu implementation
    """
    def __init__(self, profile_generator, net_id):
        """
        Initialize PopulationGenerator.
        """
        self._prof_gen = profile_generator
        self._net_id = net_id

    def header_struct(self, pop, annarchy_dir):
        """
        Generate the c-style struct definition for a population object.

        Parameters:

            pop: Population object
            annarchy_dir: working directory

        Returns:

            dictionary: include directive, pointer definition, call statements
                        and other informations needed by CodeGenerator to
                        create the ANNarchy.cpp/.cu code.

                        Please note, the returned dictionary may vary dependent
                        on population type and used parallelization paradigm.

        Exceptions:

            Will cause a NotImplementedError as it should implemented by the
            inheriting generator classes.
        """
        raise NotImplementedError

    def reset_computesum(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _delay_code(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _generate_decl_and_acc(self, pop):
        """
        Data exchange between Python and ANNarchyCore library is done for default
        attributes by accessing the C++ objects directly via the nanobind wrapper.

        See also `_generate_pop_wrapper` in `ANNarchy.generator.NanoBind.Generator`
        """
        # Parameters, Variables
        declaration, accessors, already_processed = self._generate_default_get_set(pop)

        # The conductance/current variables for spiking neurons are stored in
        # pop.neuron_type.description['variables'] but only if they are used.
        if pop.neuron_type.type == 'spike':
            try:
                all_targets = set(pop.neuron_type.description['targets'] + pop.targets)
            except TypeError:
                # The projection has multiple targets
                all_targets = set(pop.neuron_type.description['targets'] + pop.targets[0])

            for target in sorted(list(all_targets)):
                attr_name = 'g_'+target
                if attr_name not in already_processed:
                    # we assume here, that targets are local variables
                    id_dict = {
                        'type' : ConfigManager().get('precision', self._net_id),
                        'name': attr_name,
                        'attr_type': 'variable'
                    }
                    declaration += self._templates['attr_decl']['local'] % id_dict
                    already_processed.append(attr_name)

        # Global operations
        if len(pop.global_operations) != 0:
            declaration += """
    // Global operations
"""
            for op in pop.global_operations:
                op_dict = {
                    'type': ConfigManager().get('precision', self._net_id),
                    'op': op['function'],
                    'var': op['variable']
                }
                
                if _check_paradigm("openmp", self._net_id):
                    declaration += """    %(type)s _%(op)s_%(var)s;
""" % op_dict
                elif _check_paradigm("cuda", self._net_id):
                    declaration += """
    %(type)s _%(op)s_%(var)s;
    %(type)s* _gpu_%(op)s_%(var)s;
""" % op_dict
                else:
                    raise NotImplementedError

        # Arrays for the random numbers
        declaration += """
    // Random numbers
"""
        for rd in pop.neuron_type.description['random_distributions']:
            declaration += self._templates['rng'][rd['locality']]['decl'] % {
                'rd_name' : rd['name'],
                'type': rd['ctype'],
                'template': rd['template'] % {'float_prec':ConfigManager().get('precision', self._net_id)}
            }

        return declaration, accessors

    @staticmethod
    def _get_attr(pop, name):
        """
        Small helper function, used for instance in self.update_spike_neuron()
        """
        for attr in pop.neuron_type.description['variables'] + pop.neuron_type.description['parameters']:
            if attr['name'] == name:
                return attr

        return None

    @staticmethod
    def _get_attr_and_type(pop, name):
        """
        Small helper function, used for instance in self.update_spike_neuron().

        For a given variable name, the data container is searched and checked,
        whether it is a local or global variable, a random variable or a
        variable related to global operations.
        """
        for attr in pop.neuron_type.description['parameters']:
            if attr['name'] == name:
                return 'par', attr

        for attr in pop.neuron_type.description['variables']:
            if attr['name'] == name:
                return 'var', attr

        for attr in pop.neuron_type.description['random_distributions']:
            if attr['name'] == name:
                return 'rand', attr

        # the given name wasn't either an attribute nor a random distribution,
        # lets test if it was a psp
        for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
            if name == "sum("+target+")":
                return 'psp', { 'ctype': ConfigManager().get('precision', pop.net_id), 'name': '_sum_'+target }

        return None, None

    def _init_fr(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _init_globalops(self, pop):
        """
        Generate the C++ codes for the initialization of global operations
        within the Population::init_population() method.
        """
        if len(pop.global_operations) == 0:
            return ""

        code = "\n// Initialize global operations\n"
        for op in pop.global_operations:
            ids = {
                'op': op['function'],
                'var': op['variable'],
                'type': ConfigManager().get('precision', self._net_id)
            }

            if _check_paradigm("openmp", self._net_id):
                code += """_%(op)s_%(var)s = 0.0;
""" % ids

            elif _check_paradigm("cuda", self._net_id):
                code += """_%(op)s_%(var)s = 0.0;
cudaMalloc((void**)&_gpu_%(op)s_%(var)s, sizeof(%(type)s));
""" % ids

            else:
                raise NotImplementedError

        return tabify(code, 2)

    def _init_population(self, pop):
        """
        Generate the codes for the C++ function Population::init_population() method.
        """
        code = ""
        attr_tpl = self._templates['attribute_cpp_init']
        already_processed = []

        # Parameters
        for var in pop.neuron_type.description['parameters']:
            # Avoid doublons
            if var['name'] in already_processed:
                continue

            if _check_paradigm("cuda", self._net_id) and var['locality'] == "global":
                code += attr_tpl[var['locality']]['parameter'] % {'name': var['name']}
            else:
                init = 'false' if var['ctype'] == 'bool' else ('0' if var['ctype'] == 'int' else '0.0')
                var_ids = {'id': pop.id, 'name': var['name'], 'type': var['ctype'],
                        'init': init, 'attr_type': 'parameter'}
                code += attr_tpl[var['locality']] % var_ids

            already_processed.append(var['name'])

        # Variables
        for var in pop.neuron_type.description['variables']:
            # Avoid doublons
            if var['name'] in already_processed:
                continue

            init = 'false' if var['ctype'] == 'bool' else ('0' if var['ctype'] == 'int' else '0.0')
            var_ids = {'id': pop.id, 'name': var['name'], 'type': var['ctype'],
                       'init': init, 'attr_type': 'variable'}

            if _check_paradigm("cuda", self._net_id) and var['locality'] == "global":
                code += attr_tpl[var['locality']]['variable'] % var_ids
            else:
                code += attr_tpl[var['locality']] % var_ids

            already_processed.append(var['name'])

        # Random numbers 
        if _check_paradigm("openmp", self._net_id):
            if len(pop.neuron_type.description['random_distributions']) > 0:
                rng_code = "\n// Random numbers\n"
                for rd in pop.neuron_type.description['random_distributions']:
                    rng_ids = {
                        'id': pop.id,
                        'rd_name': rd['name'],
                        'type': rd['ctype'],
                    }
                    rng_code += self._templates['rng'][rd['locality']]['init'] % rng_ids
                
                code += tabify(rng_code,2)

        else:
            if len(pop.neuron_type.description['random_distributions']) > 0:
                code += """
                // Random numbers"""
                for dist in pop.neuron_type.description['random_distributions']:
                    rng_ids = {
                        'id': pop.id,
                        'rd_name': dist['name'],
                    }
                    code += self._templates['rng'][dist['locality']]['init'] % rng_ids

        # Global operations
        code += self._init_globalops(pop)

        # rate-coded targets
        if pop.neuron_type.type == 'rate':
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                ids = {
                    'id': pop.id,
                    'name': "_sum_"+target,
                    'attr_type': 'psp',
                    'type': ConfigManager().get('precision', self._net_id),
                    'init': 0.0
                }
                code += attr_tpl['local'] % ids

        # or unused synaptic spiking targets
        else:
            try:
                all_targets = set(pop.neuron_type.description['targets'] + pop.targets)
            except TypeError:
                # The projection has multiple targets
                all_targets = set(pop.neuron_type.description['targets'] + pop.targets[0])

            for target in sorted(list(all_targets)):
                attr_name = 'g_'+target
                if attr_name not in already_processed:
                    id_dict = {
                        'type' : ConfigManager().get('precision', self._net_id),
                        'name': attr_name,
                        'attr_type': 'variable',
                        'init': 0.0
                    }
                    code += self._templates['attribute_cpp_init']['local'] % id_dict
                    already_processed.append(attr_name)
                    
        return code

    def _local_functions(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _stop_condition(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _update_fr(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _update_globalops(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _update_random_distributions(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _update_rate_neuron(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _update_spiking_neuron(self, pop):
        "Implemented by child class"
        raise NotImplementedError

    def _clear_container(self, pop):
        """
        Generate code template to destroy allocated container of the C++ object *pop*. Should be implemented by child
        class.

        User defined elements, parallelization support data structures or similar are not considered. Consequently
        implementing generators should extent the resulting code template.
        """
        from ANNarchy.generator.Utils import tabify
        code = """
    #ifdef _DEBUG
        std::cout << "PopStruct%(id)s::clear()" << std::endl;
    #endif
""" % {'id': pop.id}

       # Variables
        code += "// Parameters\n"
        for attr in pop.neuron_type.description['parameters']:
            # HD: we need to clear only local parameters, the others are no vectors
            if attr['locality'] == "local":
                # HD: clear alone does not deallocate, it only resets size.
                #     So we need to call shrink_to_fit afterwards.
                ids = {'ctype': attr['ctype'], 'name': attr['name']}
                code += "%(name)s.clear();\n" % ids
                code += "%(name)s.shrink_to_fit();\n" % ids
        code+="\n"

        # Variables
        code += "// Variables\n"
        for attr in pop.neuron_type.description['variables']:
            # HD: we need to clear only local variables, the others are no vectors
            if attr['locality'] == "local":
                # HD: clear alone does not deallocate, it only resets size.
                #     So we need to call shrink_to_fit afterwards.
                ids = {'ctype': attr['ctype'], 'name': attr['name']}
                code += "%(name)s.clear();\n" % ids
                code += "%(name)s.shrink_to_fit();\n" % ids

        # Spike-specific code
        if pop.neuron_type.description['type'] == 'spike':
            code += self._templates['spike_specific']['spike']['clear']

            # Mean - FR
            code += """
// Mean Firing Rate
for (auto it = _spike_history.begin(); it != _spike_history.end(); it++) {
    while(!it->empty())
        it->pop();
}
_spike_history.clear();
_spike_history.shrink_to_fit();
"""

        # Random variables
        code += "\n// RNGs\n"
        for dist in pop.neuron_type.description['random_distributions']:
            rng_ids = {
                'id': pop.id,
                'rd_name': dist['name'],
            }
            code += self._templates['rng'][dist['locality']]['clear'] % rng_ids

        code = tabify(code, 2)
        return code

    def _size_in_bytes(self, pop):
        """
        Generate code template to determine size in bytes for the C++ object *pop*. Please note, that this contain only
        default elements (parameters, variables). User defined elements, parallelization support data structures or
        similar are not considered.

        Consequently implementing generators should extent the resulting code template. This is done by filling the
        'size_in_bytes' field in the _specific_template.
        """
        if 'size_in_bytes' in pop._specific_template.keys():
            return pop._specific_template['size_in_bytes']

        from ANNarchy.generator.Utils import tabify
        code = ""

        # Parameters
        code += "// Parameters\n"
        for attr in pop.neuron_type.description['parameters']:
            ids = {'ctype': attr['ctype'], 'name': attr['name']}
            if attr['locality'] == "global":
                code += "size_in_bytes += sizeof(%(ctype)s);\t// %(name)s\n" % ids
            else:
                code += "size_in_bytes += sizeof(std::vector<%(ctype)s>) + sizeof(%(ctype)s) * %(name)s.capacity();\t// %(name)s\n" % ids

        # Variables
        code += "// Variables\n"
        for attr in pop.neuron_type.description['variables']:
            ids = {'ctype': attr['ctype'], 'name': attr['name']}
            if attr['locality'] == "global":
                code += "size_in_bytes += sizeof(%(ctype)s);\t// %(name)s\n" % ids
            else:
                code += "size_in_bytes += sizeof(std::vector<%(ctype)s>) + sizeof(%(ctype)s) * %(name)s.capacity();\t// %(name)s\n" % ids

        # Random variables
        code +="// RNGs\n"
        if _check_paradigm("openmp", self._net_id):
            for dist in pop.neuron_type.description['random_distributions']:
                ids = {
                    'ctype': dist['ctype'],
                    'name': dist['name']
                }
                if dist['locality'] == "local":
                    code += "size_in_bytes += sizeof(std::vector<%(ctype)s>) + sizeof(%(ctype)s) * %(name)s.capacity();\t// %(name)s\n" % ids
                else:
                    code += "size_in_bytes += sizeof(%(ctype)s);\t// %(name)s\n" % ids
        else:
            for dist in pop.neuron_type.description['random_distributions']:
                code += "size_in_bytes += sizeof(curandState*);\t// gpu_%(name)s\n" % {'name': dist['name']}

        code = tabify(code, 2)
        return code

    def _generate_default_get_set(self, pop):
        """
        Generate a get/set template for all attributes in the given population
        """
        declaration = ""        # Attribute declarations
        accessors = ""          # No default accessor methods.
        already_processed = []
        code_ids_per_type = {}

        # Sort the parameters/variables per type
        for var in pop.neuron_type.description['parameters'] + pop.neuron_type.description['variables']:
            if var['name'] in already_processed:
                continue

            # add an empty list for this type if needed
            if var['ctype'] not in code_ids_per_type.keys():
                code_ids_per_type[var['ctype']] = []

            # Important which template to choose
            locality = var['locality']
            attr_type = 'parameter' if var in pop.neuron_type.description['parameters'] else 'variable'

            # For GPUs we need to tell the host that this variable need to be updated
            if _check_paradigm("cuda", self._net_id):
                if attr_type == "parameter" and locality == "global":
                    read_dirty_flag = ""
                    write_dirty_flag = ""
                else:
                    write_dirty_flag = "%(name)s_host_to_device = true;" % {'name': var['name']}
                    read_dirty_flag = "if ( %(name)s_device_to_host < t ) device_to_host();" % {'name': var['name']}
            else:
                read_dirty_flag = ""
                write_dirty_flag = ""

            # add to the processing list
            code_ids_per_type[var['ctype']].append({
                'type' : var['ctype'],
                'name': var['name'],
                'locality': locality,
                'attr_type': attr_type,
                'write_dirty_flag': write_dirty_flag,
                'read_dirty_flag': read_dirty_flag
            })

            already_processed.append(var['name'])

        # For rate-coded models add _sum_target
        if pop.neuron_type.type == "rate":
            for target in sorted(list(set(pop.neuron_type.description['targets'] + pop.targets))):
                prec_type = ConfigManager().get('precision', self._net_id)

                # add to the processing list
                code_ids_per_type[prec_type].append({
                    'type' : prec_type,
                    'name': "_sum_"+target,
                    'locality': 'local',
                    'attr_type': 'psp',
                    'write_dirty_flag': "_sum_"+target+"_host_to_device = true;",
                    'read_dirty_flag': "if ( _sum_"+target+"_device_to_host < t ) device_to_host();"
                })

        # For spiking models we add spike vector
        if pop.neuron_type.type == "spike":
            # create new type category if needed
            if "int" not in code_ids_per_type.keys():
                code_ids_per_type["int"] = []

            # add to the processing list
            code_ids_per_type["int"].append({
                'type' : "int",
                'name': "spiked",
                'locality': 'local',
                'attr_type': 'spike',
                'write_dirty_flag': "",
                'read_dirty_flag': ""
            })

        # Generate code snippets
        for ctype in code_ids_per_type.keys():
            for ids in code_ids_per_type[ctype]:
                locality = ids['locality']

                # Declaration codes
                if ids['name'] == "spiked":
                    declaration += ""   # already declared
                elif _check_paradigm("cuda", self._net_id) and locality == "global":
                    declaration += self._templates['attr_decl'][locality][ids['attr_type']] % ids
                else:
                    declaration += self._templates['attr_decl'][locality] % ids

        return declaration, accessors, already_processed


"""
This file is part of ANNarchy.

:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""
import unittest
from ANNarchy import setup

setup(paradigm="cuda")

from Common import *
from Interface import *
from Neuron import *
from RateSynapse import *
from SpikeSynapse import *


if __name__ == '__main__':
    unittest.main(verbosity=2)



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Synapse import Synapse

class SharedSynapse(Synapse):
    """
    Shared synapse for report()
    """
    # For reporting
    _instantiated = []

    def __init__(self, psp, operation, name="Shared synapse", description="Weight shared over all synapses of the projection."):
        """
        """
        # Shared synapses are non-plastic.
        Synapse.__init__(self, 
            psp=psp, operation=operation,
            name=name, 
            description=description
        )

        # For reporting
        self._instantiated.append(True)


# distutils: language = c++

import numpy as np
cimport numpy as np

from ANNarchy.intern.ConfigManagement import get_global_config

cpdef np.ndarray raster_plot(list data):
    """
    Transforms recorded spikes into a (N, 2) array to easily display the raster plot.
    """
    cdef int N, n, t
    cdef np.ndarray res
    cdef list d
    N = len(data)
    d = []
    for n in xrange(N):
        for t in data[n]:
            d.append([t, n])
    res = np.array(d)
    return res

cpdef np.ndarray smoothed_rate(dict data, float smooth):
    """ Takes the recorded spikes of a population and returns a smoothed firing rate.

    Parameters:

    * *data* the dictionary returned by ``Pop.get_record()['spike']``

    * *smooth* the smoothing time constant (default: 0 ms)
    """
    cdef np.ndarray res
    cdef int N, d
    cdef int n, t, timing, last_spike, idx
    cdef float dt, delta
    cdef list spikes

    # Retrieve simulation time step
    dt = get_global_config('dt')

    # Number of neurons
    d = data['stop'] - data['start']
    N = len(data['data'])

    # Prepare the matrix
    rates = np.zeros((N, d))

    # Compute instantaneous firing rate
    idx = 0
    for n, spikes in data['data'].items():
        last_spike = data['start'] - int(100.0/dt)
        for timing in spikes:
            if last_spike>data['start']:
                rates[idx, last_spike:timing] = 1000.0/dt/float(timing - last_spike)  
            last_spike = timing 
        idx += 1

    if smooth == 0.0:
        return rates

    smoothed_rate = np.zeros((N, d))
    smoothed_rate[:, 0] = rates[:, 0]
    delta = dt/smooth
    for t in xrange(d-1):
        smoothed_rate[:, t+1] = smoothed_rate[:, t] + (rates[:, t+1] - smoothed_rate[:, t])*delta
    return smoothed_rate

cpdef np.ndarray population_rate(dict data, float smooth):
    """ Takes the recorded spikes of a population and returns a smoothed firing rate for the whole population.

    Parameters:

    * *data* the dictionary returned by ``Pop.get_record()['spike']``

    * *smooth* the smoothing time constant (default: dt)
    """
    cdef np.ndarray res
    cdef int N, d
    cdef int n, t, timing, last_spike
    cdef float dt, delta
    cdef list spikes

    # Retrieve simulation time step
    dt = get_global_config('dt')

    # Number of neurons
    d = data['stop'] - data['start'] + 1
    N = len(data['data'])

    # Prepare the matrix
    rates = np.zeros(d)

    # Compute histogram
    for n, spikes in data['data'].items():
        for t in spikes:
            rates[t - data['start']] += 1
    rates /= dt*N/1000.0
    
    # print(smooth, dt)

    if smooth <= dt:
        return rates

    smoothed_rate = np.zeros(d)
    smoothed_rate[0] = rates[0]
    delta = dt/smooth
    for t in xrange(d-1):
        smoothed_rate[t+1] = smoothed_rate[t] + (rates[t+1] - smoothed_rate[t])*delta

    return smoothed_rate




"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Neuron import Neuron

class BoldModel(Neuron):
    r"""
    Base class to define a BOLD model to be used in a BOLD monitor.

    A BOLD model is quite similar to a regular rate-coded neuron. It gets a weighted sum of inputs with a specific target (e.g. I_CBF) and compute a single output variable (called `BOLD` in the predefined models, but it could be anything).

    The main difference is that a BOLD model should also declare which targets are used for the input signal:

    ```python
    bold_model = BoldModel(
        parameters = dict(
            tau = 1000.
        ),
        equations = [
            'I_CBF = sum(I_CBF)',
            # ...
            'tau * dBOLD/dt = I_CBF - BOLD',
        ],
        inputs = "I_CBF"
    )
    ```

    The provided BOLD models follow the Balloon model (Buxton et al., 1998) with the different variations studied in (Stephan et al., 2007). Those models all compute the vascular response to neural activity through a dampened oscillator:

    $$
        \frac{ds}{dt} = \phi \, I_\text{CBF} - \kappa \, s - \gamma \, (f_{in} - 1)
    $$

    $$
        \frac{df_{in}}{dt} = s
    $$

    This allows to compute the oxygen extraction fraction:

    $$
        E = 1 - (1 - E_{0})^{ \frac{1}{f_{in}} }
    $$

    The (normalized) venous blood volume is computed as:

    $$
        \tau_0 \, \frac{dv}{dt} = (f_{in} - f_{out})
    $$

    $$
        f_{out} = v^{\frac{1}{\alpha}}
    $$

    The level of deoxyhemoglobin into the venous compartment is computed by:

    $$
        \tau_0 \, \frac{dq}{dt} = f_{in} \, \frac{E}{E_0} - \frac{q}{v} \, f_{out}
    $$

    Using the two signals $v$ and $q$, there are two ways to compute the corresponding BOLD signal:

    * **N:** Non-linear BOLD equation:

    $$
        BOLD = v_0 \, ( k_1 \, (1-q) + k_2 \, (1- \dfrac{q}{v}) + k_3 \, (1 - v) )
    $$

    * **L:** Linear BOLD equation:

    $$
        BOLD = v_0 \, ((k_1 + k_2) \, (1 - q) + (k_3 - k_2) \, (1 - v)) 
    $$

    Additionally, the three coefficients $k_1$, $k_2$, $k_3$ can be computed in two different ways:

    * **C:** classical coefficients from (Buxton et al., 1998):

    $$k_1            = (1 - v_0) \, 4.3 \, v_0 \, E_0 \, \text{TE}$$

    $$k_2            = 2 \, E_0$$

    $$k_3            = 1 - \epsilon$$

    * **R:** revised coefficients from (Obata et al., 2004):

    $$k_1            = 4.3 \, v_0 \, E_0 \, \text{TE}$$

    $$k_2            = \epsilon \, r_0 \, E_0 \, \text{TE}$$

    $$k_3            = 1 - \epsilon$$

    This makes a total of four different BOLD model (`balloon_RN`, `balloon_RL`, `balloon_CN`, `balloon_CL`) which are provided by the extension. The different parameters can be modified in the constructor. Additionally, we also provide the model that was used in (Maith et al., 2021) and the two-inputs model of (Maith et al, 2022).
    
    > Buxton, R. B., Wong, E. C., and Frank, L. R. (1998). Dynamics of blood flow and oxygenation changes during brain activation: the balloon model. Magnetic resonance in medicine 39, 855-864. doi:10.1002/mrm.1910390602
    
    > Friston et al. (2000). Nonlinear responses in fMRI: the balloon model, volterra kernels, and other hemodynamics. NeuroImage 12, 466-477
    
    > Buxton et al. (2004). Modeling the hemodynamic response to brain activation. Neuroimage 23, 220-233. doi:10.1016/j.neuroimage.2004.07.013
    
    > Stephan et al. (2007). Comparing hemodynamic models with DCM. Neuroimage 38, 387-401. doi:10.1016/j.neuroimage.2007.07.040
    
    > Maith et al. (2021) A computational model-based analysis of basal ganglia pathway changes in Parkinson's disease inferred from resting-state fMRI. European Journal of Neuroscience. 2021; 53: 2278-2295. doi:10.1111/ejn.14868 
    
    > Maith et al. (2022) BOLD monitoring in the neural simulator ANNarchy. Frontiers in Neuroinformatics 16. doi:10.3389/fninf.2022.790966.

    :param parameters: parameters of the model and their initial value.
    :param equations: equations defining the temporal evolution of variables.
    :param inputs: single variable or list of input signals (e.g. 'I_CBF' or ['I_CBF', 'I_CMRO2']).
    :param output: output variable of the model (default is 'BOLD').
    :param name: optional model name.
    :param description: optional model description.
    """
    def __init__(self, parameters, equations, inputs, output=["BOLD"], name="Custom BOLD model", description=""):

        # The processing in BoldMonitor expects lists, but the interface
        # should allow also single strings (if only one variable is considered)
        self._inputs = [inputs] if isinstance(inputs, str) else inputs
        self._output = [output] if isinstance(output, str) else output

        Neuron.__init__(self, parameters=parameters, equations=equations, name=name, description=description)
        
        self._model_instantiated = False    # activated by BoldMonitor



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Population import Population
from ANNarchy.core.Neuron import Neuron
from ANNarchy.core import Global

from ANNarchy.intern.ConfigManagement import ConfigManager
from ANNarchy.intern import Messages

class Spike2RatePopulation(Population):
    """
    Converts a population of spiking neurons into a population of rate-coded neurons.

    When building a hybrid network, one may need to convert spike trains into an instantaneous firing. Creating a ``Spike2RatePopulation`` allows to get a rate-coded population of the same size as the spiking population.

    Two modes are available:

    * ``window``: at each each time step t, the number of spikes in the last T milliseconds (defined by the parameter ``window``) is counted and divided by the interval.

    * ``isi``: the inter-spike interval is computed for each new spike, and filtered out through an asymmetrical kernel function. 


    The firing rate ``r`` of each neuron represents by default the firing rate in Hz. The output can be scaled with the parameter ``scaling``. For example, if you want that ``r = 1.0`` represents a firing rate of 100Hz, you can set ``scaling`` to 100.0. The default is 1.0.

    By definition, the firing rate varies abruptly each time a new spike is perceived. The output can be smoothed with a low-pass filter of time constant ``smooth``.

    .. code-block:: python

        from ANNarchy.extensions.hybrid import Spike2RatePopulation

        pop2 = Spike2RatePopulation(
            population=pop1, 
            name='rate-coded', 
            window=50.0, 
            smooth=100.0, 
            scaling=100.0
        )
    """
    def __init__(self, population, name=None, mode='window', window = 100.0, scaling=1.0, smooth=1.0, cut=3.0, copied=False):
        """
        :param population: the Population to convert. Its neuron type must be rate-coded.
        :param name: the (optional) name of the hybrid population.
        :param mode: mode of computation of the firing rate. ``'window'`` (default) or ``'isi'``.
        :param window: the extent of the sliding window (in ms) used to compute the firing rate in the 'window' mode(default: 100.0 ms).
        :param cut: cutting frequency of the ``'isi'`` kernel. default 2.0.
        :param scaling: the scaling of the firing rate. Defines what a firing rate of 1 Hz outputs (default: 1.0).
        :param smooth: time constant (in ms) of the low-pass filter used to smooth the firing rate (default: 1 ms, i.e no smoothing)
        """
        self.population = population
        self.name = name
        self.mode = mode
        self.scaling = scaling
        self.window = window
        self.smooth = smooth
        self.cut = cut
        self.copied = copied

        if not self.population.neuron_type.description['type'] == 'spike':
            Messages._error('the population ' + self.population.name + ' must contain spiking neurons.')
            
        if self.mode == 'window':
            self._code = self._create_window()
        elif self.mode == 'adaptive':
            self._code = self._create_adaptive()
        elif self.mode == 'isi':
            self._code = self._create_isi()
        else:
            Messages._error('Spike2RatePopulation: Unknown method ' + self.mode)
            

        self._specific = True

    def _copy(self):
        "Returns a copy of the population when creating networks. Internal use only."
        return Spike2RatePopulation(population=self.population, name=self.name, mode=self.mode, window=self.window, scaling=self.scaling, smooth=self.smooth, cut=self.cut, copied=True)
    
    def generate(self):
        """
        return the corresponding code template
        """
        return self._code

    def _create_isi(self):

        # Create the description, but it will not be used for generation
        Population.__init__(
            self, 
            geometry = self.population.geometry, 
            name=self.name, 
            neuron = Neuron(
                parameters="""
                    cut = %(cut)s : population
                    scaling = %(scaling)s : population
                    smooth = %(smooth)s : population
                """ % {'cut': self.cut, 'scaling': self.scaling, 'smooth': self.smooth} ,
                equations="r = 0.0"
            ),
            copied=self.copied
        )

        # Generate specific code
        omp_code = "#pragma omp for" if ConfigManager().get('num_threads', self.net_id) > 1 else ""
        code = """#pragma once

#include "pop%(id_pre)s.hpp"
extern PopStruct%(id_pre)s pop%(id_pre)s;

struct PopStruct%(id)s{
    // Number of neurons
    int size;
    bool _active;

    // Global parameter cut
    %(float_prec)s cut;

    // Global parameter scaling
    %(float_prec)s  scaling;

    // Global parameter smooth
    %(float_prec)s  smooth ;

    // Local variable r
    std::vector< %(float_prec)s > r ;

    // Store the last spike
    std::vector< long int > last_spike;
    std::vector< %(float_prec)s > isi;
    std::vector< %(float_prec)s > support;

    int get_size() { return size; }
    void set_size(int value) { size = value; }

    bool is_active() { return _active; }
    void set_active(bool value) { _active = value; }

    %(float_prec)s get_cut() { return cut; }
    void set_cut(%(float_prec)s value) { cut = value; }

    %(float_prec)s get_scaling() { return scaling; }
    void set_scaling(%(float_prec)s value) { scaling = value; }

    %(float_prec)s get_smooth() { return smooth; }
    void set_smooth(%(float_prec)s value) { smooth = value; }

    std::vector< %(float_prec)s > get_r() { return r; }
    void set_r(std::vector< %(float_prec)s > value) { r = value; }
    %(float_prec)s get_single_r(int rk) { return r[rk]; }
    void set_single_r(int rk, %(float_prec)s value) { r[rk] = value; }

    void init_population() {
        size = %(size)s;
        _active = true;

        last_spike = std::vector< long int >(size, -10000L);
        isi = std::vector< %(float_prec)s >(size, 10000.0);
        support = std::vector< %(float_prec)s >(size, 10000.0);
    }

    void update_rng() {

    }

    void update() {
        // Updating the local variables of Spike2Rate population %(id)s
        if(_active){
            %(omp_code)s
            for(int i = 0; i < size; i++){
                // Increase when spiking
                if (pop%(id_pre)s.last_spike[i] == t-1){
                   isi[i] = %(float_prec)s(t-1 - last_spike[i]);
                   support[i] = %(float_prec)s(t-1 - last_spike[i]);
                   last_spike[i] = t-1;
                }
                else if( %(float_prec)s(t - last_spike[i]) <= isi[i]){
                        // do nothing
                }
                else if( %(float_prec)s(t - last_spike[i]) <= cut*isi[i]){
                        support[i] += 1.0 ;
                }
                else{
                    support[i] = 10000.0 ;
                }

                r[i] += dt*(1000.0/scaling/support[i]/dt - r[i])/smooth;
            }
        }
    }
};
""" % {'id' : self.id, 'id_pre': self.population.id, 'omp_code': omp_code, 'size': self.size, 'float_prec': ConfigManager().get('precision', self.net_id) }

        return code

    def _create_window(self):

        # Create the description, but it will not be used for generation
        Population.__init__(
            self, 
            geometry = self.population.geometry, 
            name=self.name, 
            neuron = Neuron(
                parameters="""
                    window = %(window)s : population
                    scaling = %(scaling)s : population
                    smooth = %(smooth)s : population
                """ % {'window': self.window, 'scaling': self.scaling, 'smooth': self.smooth} ,
                equations="r = 0.0"
            ) ,
            copied=self.copied
        )

        # Generate specific code
        omp_code = "#pragma omp for" if ConfigManager().get('num_threads', self.net_id) > 1 else ""
        code = """#pragma once

#include "pop%(id_pre)s.hpp"
extern PopStruct%(id_pre)s pop%(id_pre)s;

struct PopStruct%(id)s{
    // Number of neurons
    int size;
    bool _active;

    // Global parameter window
    %(float_prec)s  window ;

    // Global parameter scaling
    %(float_prec)s  scaling ;

    // Global parameter smooth
    %(float_prec)s  smooth ;

    // Local variable r
    std::vector< %(float_prec)s > r ;

    // Store the last spikes
    std::vector< std::vector<long int> > last_spikes;

    int get_size() { return size; }
    void set_size(int value) { size = value; }

    bool is_active() { return _active; }
    void set_active(bool value) { _active = value; }

    %(float_prec)s get_window() { return window; }
    void set_window(%(float_prec)s value) { window = value; }

    %(float_prec)s get_scaling() { return scaling; }
    void set_scaling(%(float_prec)s value) { scaling = value; }

    %(float_prec)s get_smooth() { return smooth; }
    void set_smooth(%(float_prec)s value) { smooth = value; }

    std::vector< %(float_prec)s > get_r() { return r; }
    void set_r(std::vector< %(float_prec)s > value) { r = value; }
    %(float_prec)s get_single_r(int rk) { return r[rk]; }
    void set_single_r(int rk, %(float_prec)s value) { r[rk] = value; }

    void init_population() {
        size = %(size)s;
        _active = true;

        last_spikes = std::vector< std::vector<long int> >(size, std::vector<long int>());
    }

    void update_rng() {

    }

    void update() {
        // Updating the local variables of Spike2Rate population %(id)s
        if(_active){
            int pop%(id)s_nb, pop%(id)s_out;
            %(omp_code)s
            for(int i = 0; i < size; i++){
                // Increase when spiking
                if (pop%(id_pre)s.last_spike[i] == t-1){
                   last_spikes[i].push_back(t-1);
                }
                pop%(id)s_nb = 0;
                pop%(id)s_out = -1;
                for(int j=0; j < last_spikes[i].size(); j++){
                    if(last_spikes[i][j] >= t -1 - (long int)(window/dt) ){
                        pop%(id)s_nb++;
                    }
                    else{
                        pop%(id)s_out = j;
                    }
                }

                if (pop%(id)s_out > -1){
                    last_spikes[i].erase(last_spikes[i].begin());
                }

                r[i] += dt*(1000.0/scaling / window * %(float_prec)s(pop%(id)s_nb) - r[i] ) / smooth;
            }
        }
    }
};
""" % {'id' : self.id, 'id_pre': self.population.id, 'omp_code': omp_code, 'size': self.size, 'float_prec': ConfigManager().get('precision', self.net_id)}

        return code

    def _create_adaptive(self):

        # Create the description, but it will not be used for generation
        Population.__init__(
            self, 
            geometry = self.population.geometry, 
            name=self.name, 
            neuron = Neuron(
                parameters="""
                    window = %(window)s : population
                    scaling = %(scaling)s : population
                    smooth = %(smooth)s : population
                """ % {'window': self.window, 'scaling': self.scaling, 'smooth': self.smooth} ,
                equations="r = 0.0"
            ) ,
            copied=self.copied
        )

        # Generate specific code
        omp_code = "#pragma omp for private(pop%(id)s_nb, pop%(id)s_out)" if ConfigManager().get('num_threads', self.net_id) > 1 else ""
        code = """#pragma once

#include "pop%(id_pre)s.hpp"
extern PopStruct%(id_pre)s pop%(id_pre)s;

struct PopStruct%(id)s{
    // Number of neurons
    int size;
    bool _active;

    // Local parameter window
    %(float_prec)s  window ;
    std::vector< %(float_prec)s >  ad_window ;

    // Global parameter scaling
    %(float_prec)s  scaling ;

    // Global parameter smooth
    %(float_prec)s  smooth ;

    // Local variable r
    std::vector< %(float_prec)s > r ;
    std::vector< std::vector< %(float_prec)s > > recorded_r ;
    bool record_r ;

    // Store the last spikes
    std::vector< std::vector<long int> > last_spikes;
    std::vector< %(float_prec)s > isi;

    int get_size() { return size; }
    void set_size(int value) { size = value; }

    bool is_active() { return _active; }
    void set_active(bool value) { _active = value; }

    %(float_prec)s get_window() { return window; }
    void set_window(%(float_prec)s value) { window = value; }

    %(float_prec)s get_scaling() { return scaling; }
    void set_scaling(%(float_prec)s value) { scaling = value; }

    %(float_prec)s get_smooth() { return smooth; }
    void set_smooth(%(float_prec)s value) { smooth = value; }

    std::vector< %(float_prec)s > get_r() { return r; }
    void set_r(std::vector< %(float_prec)s > value) { r = value; }
    %(float_prec)s get_single_r(int rk) { return r[rk]; }
    void set_single_r(int rk, %(float_prec)s value) { r[rk] = value; }

    void init_population() {
        size = %(size)s;
        _active = true;

        last_spikes = std::vector< std::vector<long int> >(size, std::vector<long int>());
        ad_window = std::vector< %(float_prec)s >(size, window);
        isi = std::vector< %(float_prec)s >(size, 10000.0);
    }

    void update_rng() {

    }

    void update() {
        // Updating the local variables of Spike2Rate population %(id)s
        if(_active){
            int pop%(id)s_nb, pop%(id)s_out;
            %(omp_code)s
            for(int i = 0; i < size; i++){
                // Increase when spiking
                if (pop%(id_pre)s.last_spike[i] == t-1){
                    if(last_spikes[i].size() > 0)
                        isi[i] = %(float_prec)s(t-1 - last_spikes[i][last_spikes[i].size()-1]);
                    else
                        isi[i] = 10000.0;
                    last_spikes[i].push_back(t-1);
                }
                pop%(id)s_nb = 0;
                pop%(id)s_out = -1;
                for(int j=0; j < last_spikes[i].size(); j++){
                    if(last_spikes[i][j] >= t -1 - (long int)(ad_window[i]/dt) ){
                        pop%(id)s_nb++;
                    }
                    else{
                        pop%(id)s_out = j;
                    }
                }

                if (pop%(id)s_out > -1){
                    last_spikes[i].erase(last_spikes[i].begin(), last_spikes[i].begin()+pop%(id)s_out);
                }

                r[i] += dt*(1000.0/scaling / ad_window[i] * %(float_prec)s(pop%(id)s_nb) - r[i] ) / smooth;

                //ad_window[i] = clip(5.0*isi[i], 20.0*dt, window) ;
                ad_window[i] += dt * (clip(5.0*isi[i], 20.0*dt, window) - ad_window[i])/100.0;
                if (i==0)
                    std::cout << ad_window[i] << " " << isi[i] << std::endl;
            }
        }
    }
};
""" % {'id' : self.id, 'id_pre': self.population.id, 'omp_code': omp_code, 'size': self.size, 'float_prec': ConfigManager().get('precision', self.net_id)}

        return code

class Rate2SpikePopulation(Population):
    """
    Converts a population of rate-coded neurons into a population of spiking neurons.

    This class allows to generate spike trains based on the computations of a rate-coded network (for example doing visual pre-processing). Creating a ``Rate2SpikePopulation`` allows to get a spiking population of the same size as the rate-coded population.

    The firing rate ``r`` of the rate-coded population represents by default the desired firing rate in Hz. This value can be scaled with the parameter ``scaling``. For example, if you want that ``r = 1.0`` represents a firing rate of 100Hz, you can set ``scaling`` to 100.0.

    .. code-block:: python

        from ANNarchy.extensions.hybrid import Rate2SpikePopulation
        pop2 = Rate2SpikePopulation(
            population=pop1, 
            name='spiking', 
            scaling=100.0
        )
    """
    def __init__(self, population, name=None, scaling=1.0, refractory=None, copied=False):
        """
        :param population: the Population to convert. Its neuron type must be spiking.
        :param name: the (optional) name of the hybrid population.
        :param scaling: the scaling of the firing rate. Defines what a rate ``r`` of 1.0 means in Hz (default: 1.0).
        :param refractory: a refractory period in ms to ensure the ISI is not too high (default: None)
        """
        self.population = population
        self.refractory_init = refractory

        if not self.population.neuron_type.description['type'] == 'rate':
            Messages._error('the population ' + self.population.name + ' must contain rate-coded neurons.')
            

        # Create the description, but it will not be used for generation
        Population.__init__(
            self, 
            geometry = self.population.geometry, 
            name=name, 
            neuron = Neuron(
                parameters="""
                    scaling = %(scaling)s : population
                """ % {'scaling': scaling} ,
                equations="""
                    p = Uniform(0.0, 1.0)
                    rates = p
                """,
                spike="rates>p",
                refractory=refractory
            ) ,
            copied=self.copied
        )

        self._specific = True

    def _copy(self):
        "Returns a copy of the population when creating networks. Internal use only."
        return Rate2SpikePopulation(population=self.population, name=self.name, scaling=self.scaling, refractory=self.refractory_init, copied=True)

    def generate(self):
        omp_code = "#pragma omp for" if ConfigManager().get('num_threads', self.net_id) > 1 else ""
        omp_critical = "#pragma omp critical" if ConfigManager().get('num_threads', self.net_id) > 1 else ""

        # Generate the code
        code = """#pragma once

#include "pop%(id_pre)s.hpp"
extern PopStruct%(id_pre)s pop%(id_pre)s;

struct PopStruct1{
    // Number of neurons
    int size;
    bool _active;

    // Global parameter scaling
    %(float_prec)s  scaling ;

    // Local variable p
    std::vector< %(float_prec)s > p;

    // Local variable rates
    std::vector< %(float_prec)s > rates;

    // Random numbers (STL implementation)
    std::vector< %(float_prec)s > rand_0;
    std::uniform_real_distribution< %(float_prec)s > dist_rand_0;

    // Spiking events
    std::vector<long int> last_spike;
    std::vector<int> spiked;
    std::vector<int> refractory;
    std::vector<int> refractory_remaining;
    bool record_spike;
    std::vector<std::vector<long> > recorded_spike;

    int get_size() { return size; }
    bool is_active() { return _active; }
    bool set_active(bool val) { _active = val; }

    // Global parameter scaling
    %(float_prec)s get_scaling() { return scaling; }
    void set_scaling(%(float_prec)s val) { scaling = val; }

    // Local variable p
    std::vector< %(float_prec)s > get_p() { return p; }
    %(float_prec)s get_single_p(int rk) { return p[rk]; }
    void set_p(std::vector< %(float_prec)s > val) { p = val; }
    void set_single_p(int rk, %(float_prec)s val) { p[rk] = val; }

    // Local variable rates
    std::vector< %(float_prec)s > get_rates() { return rates; }
    %(float_prec)s get_single_rates(int rk) { return rates[rk]; }
    void set_rates(std::vector< %(float_prec)s > val) { rates = val; }
    void set_single_rates(int rk, %(float_prec)s val) { rates[rk] = val; }

    void init_population() {
        size = %(size)s;
        _active = true;

        refractory = std::vector<int>(size, 0);
        spiked = std::vector<int>(0, 0);
        last_spike = std::vector<long int>(size, -10000L);
        refractory_remaining = std::vector<int>(size, 0);

        rand_0 = std::vector< %(float_prec)s >(size, 0.0);
        dist_rand_0 = std::uniform_real_distribution< %(float_prec)s >(0.0, 1.0);
    }

    void update_rng() {
        for(int i = 0; i < size; i++)
        {
            rand_0[i] = dist_rand_0(rng);
        }
    }

    void update() {
        // Updating the local variables of population %(id)s (Rate2SpikePopulation)
        if(_active){
            spiked.clear();
            %(omp_code)s
            for(int i = 0; i < size; i++){
                rates[i] = pop%(id_pre)s.r[i] * scaling;

                if(refractory_remaining[i] > 0){ // Refractory period
                    refractory_remaining[i]--;
                }
                else if(rates[i] > rand_0[i]*1000.0/dt){
                    %(omp_critical)s
                    {
                        spiked.push_back(i);
                    }
                    last_spike[i] = t;
                    refractory_remaining[i] = refractory[i];
                }
            }
        }
    }
};
""" % {'id' : self.id, 'id_pre': self.population.id, 'omp_code': omp_code, 'omp_critical': omp_critical, 'size': self.size, 'float_prec': ConfigManager().get('precision', self.net_id) }

        return code


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

attribute_decl = {
    'local':
"""
    // Local %(attr_type)s %(name)s
    std::vector< %(type)s > %(name)s;
""",
    'semiglobal':
"""
    // Semiglobal %(attr_type)s %(name)s
    std::vector< %(type)s >  %(name)s ;
""",
    'global':
"""
    // Global %(attr_type)s %(name)s
    %(type)s  %(name)s ;
"""
}

attribute_cpp_init = {
    'local':
"""
        // Local %(attr_type)s %(name)s
        %(name)s = init_matrix_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'semiglobal':
"""
        // Semiglobal %(attr_type)s %(name)s
        %(name)s = init_vector_variable<%(type)s>(static_cast<%(type)s>(%(init)s));
""",
    'global':
"""
        // Global %(attr_type)s %(name)s
        %(name)s = %(init)s;
"""
}

attribute_cpp_size = {
    'local': """
        // Local %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();       
""",
    'semiglobal': """
        // Semiglobal %(attr_type)s %(name)s
        size_in_bytes += sizeof(std::vector<%(ctype)s>);
        size_in_bytes += sizeof(%(ctype)s) * %(name)s.capacity();
""",
    'global': """
        // Global
        size_in_bytes += sizeof(%(ctype)s);
"""
}

attribute_cpp_delete = {
    'local': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'semiglobal': """
        // %(name)s
        %(name)s.clear();
        %(name)s.shrink_to_fit();
""",
    'global': ""
}

delay = {
    # A single value for all synapses
    'uniform': {
        'declare': """
    // Uniform delay
    int delay;

    int get_delay() { return delay; }
    int get_dendrite_delay(int idx) { return delay; }
    void set_delay(int delay) { this->delay = delay; }
""",
        'init': """
    delay = delays[0][0];
"""
    },
    # An individual value for each synapse
    'nonuniform_rate_coded': {
        'declare': """
    // Non-uniform delay
    std::vector<int> delay;
    int max_delay;

    std::vector<std::vector<int>> get_delay() { return get_matrix_variable_all<int>(delay); }
    void set_delay(std::vector<std::vector<int>> value) { update_matrix_variable_all<int>(delay, value); }
    std::vector<int> get_dendrite_delay(int lil_idx) { return get_matrix_variable_row<int>(delay, lil_idx); }
    int get_max_delay() { return max_delay; }
    void set_max_delay() { this->max_delay = max_delay; }
""",
        'init': """
    delay = init_matrix_variable<int>(1);
    update_matrix_variable_all<int>(delay, delays);

    max_delay = %(pre_prefix)smax_delay;
""",
        'reset': ""
    },
    # An individual value for each synapse and a
    # buffer for spike events    
    'nonuniform_spiking': {
        'declare': """
    std::vector<int> delay;
    int max_delay;
    int idx_delay;
    std::vector< std::vector< std::vector< int > > > _delayed_spikes;

    std::vector<std::vector<int>> get_delay() { return get_matrix_variable_all<int>(delay); }
    void set_delay(std::vector<std::vector<int>> value) { update_matrix_variable_all<int>(delay, value); }
    std::vector<int> get_dendrite_delay(int lil_idx) { return get_matrix_variable_row<int>(delay, lil_idx); }
    int get_max_delay() { return max_delay; }
    void set_max_delay() { this->max_delay = max_delay; }
""",
        'init': """
    delay = init_matrix_variable<int>(1);
    update_matrix_variable_all<int>(delay, delays);

    idx_delay = 0;
    max_delay = %(pre_prefix)smax_delay;
""",
        'reset': """
        while(!_delayed_spikes.empty()) {
            auto elem = _delayed_spikes.back();
            elem.clear();
            _delayed_spikes.pop_back();
        }

        idx_delay = 0;
        max_delay =  %(pre_prefix)smax_delay ;
        _delayed_spikes = std::vector< std::vector< std::vector< int > > >(max_delay, std::vector< std::vector< int > >(post_rank.size(), std::vector< int >()) );        
"""
    }   
}

event_driven = {
    'declare': """
    std::vector<long> _last_event;
""",
    'cpp_init': """
    _last_event = init_matrix_variable<long>(-10000);
""",
    'pyx_struct': """
        vector[vector[long]] _last_event
""",
}

csr_summation_operation = {
    'sum' : """
%(pre_copy)s

const %(size_type)s * __restrict__ row_ptr = row_begin_.data();
const %(idx_type)s * __restrict__ col_idx = col_idx_.data();

for(auto it = post_ranks_.cbegin(); it != post_ranks_.cend(); it++) {
    %(idx_type)s rk_post = *it;

    sum = 0.0;
    for(%(size_type)s j = row_ptr[rk_post]; j < row_ptr[rk_post+1]; j++) {
        sum += %(psp)s;
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum;
} 
""",
    'max': """
%(pre_copy)s

const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
const %(idx_type)s* __restrict__ col_idx = col_idx_.data();

for(auto it = post_ranks_.cbegin(); it != post_ranks_.cend(); it++) {
    %(idx_type)s rk_post = *it;

    %(size_type)s j = _row_ptr[rk_post];
    sum = %(psp)s ;
    for(j = _row_ptr[rk_post]+1; j < _row_ptr[rk_post+1]; j++){
        if(%(psp)s > sum){
            sum = %(psp)s ;
        }
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum;
}
""",
    'min': """
%(pre_copy)s

const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
const %(idx_type)s* __restrict__ col_idx = col_idx_.data();

for(auto it = post_ranks_.cbegin(); it != post_ranks_.cend(); it++) {
    %(idx_type)s rk_post = *it;
    
    %(size_type)s j= _row_ptr[rk_post];
    sum = %(psp)s ;
    for(j = _row_ptr[rk_post]+1; j < _row_ptr[rk_post+1]; j++){
        if(%(psp)s < sum){
            sum = %(psp)s ;
        }
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum;
}
""",
    'mean': """
%(pre_copy)s

const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
const %(idx_type)s* __restrict__ col_idx = col_idx_.data();

for(auto it = post_ranks_.cbegin(); it != post_ranks_.cend(); it++) {
    %(idx_type)s rk_post = *it;

    sum = 0.0 ;
    for(%(size_type)s j = _row_ptr[rk_post]; j < _row_ptr[rk_post+1]; j++){
        sum += %(psp)s ;
    }
    %(post_prefix)s_sum_%(target)s%(post_index)s += sum / static_cast<%(float_prec)s>(pre_rank[i].size());
}
"""
}

###############################################################################
# Optimized kernel for default rate-coded continuous transmission using
# SIMD instructions and single weight value for all synapses in the projection.
#
# The default psp-formula:
#
#  psp = sum_(i=0)^C w * r_i
#
# can be rewritten as
#
#  psp = w * sum_(i=0)^C r_i
#
# so we can save C multiplications. Please note, this can lead to small
# deviations, but they appear to be close to the precision border
# (e. g. ~10^-17 for double)
#
# HD: for this code also lesser SSE might be suitable, but it's outdated anyways ...
###############################################################################
continuous_transmission_sse_single_weight = {
    'sum' : {
        'double': """
    #ifdef __SSE4_1__
        if (_transmission && %(post_prefix)s_active) {
            const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
            const %(idx_type)s* __restrict__ _idx = col_idx_.data();

            double _tmp_sum[2];
            double* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];
                %(size_type)s _s = row_ptr[rk_post];
                %(size_type)s _stop = row_ptr[rk_post+1];
                __m128d _tmp_reg_sum = _mm_setzero_pd();

                for (; (_s+8) < _stop; _s+=8) {
                    __m128d _tmp_r = _mm_set_pd(_pre_r[_idx[_s+1]], _pre_r[_idx[_s]]);
                    __m128d _tmp_r2 = _mm_set_pd(_pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]]);
                    __m128d _tmp_r3 = _mm_set_pd(_pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]);
                    __m128d _tmp_r4 = _mm_set_pd(_pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]]);

                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _tmp_r);
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _tmp_r2);
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _tmp_r3);
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _tmp_r4);
                }
                _mm_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #endif
""",
        'float': """
    #ifdef __SSE4_1__
        if (_transmission && %(post_prefix)s_active) {
            const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
            const %(idx_type)s* __restrict__ _idx = col_idx_.data();

            float _tmp_sum[4];
            float* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];
                %(size_type)s _s = row_ptr[rk_post];
                %(size_type)s _stop = row_ptr[rk_post+1];
                __m128 _tmp_reg_sum = _mm_setzero_ps();

                for (; (_s+16) < _stop; _s+=16) {
                    __m128 _tmp_r = _mm_set_ps(_pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]);
                    __m128 _tmp_r2 = _mm_set_ps(_pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]);
                    __m128 _tmp_r3 = _mm_set_ps(_pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]]);
                    __m128 _tmp_r4 = _mm_set_ps(_pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]]);

                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _tmp_r);
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _tmp_r2);
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _tmp_r3);
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _tmp_r4);
                }
                _mm_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with SSE4-1 support. Please check your compiler flags ..." << std::endl;
    #endif
"""
    }
}

continuous_transmission_avx_single_weight = {
    'sum' : {
        'double': """
    #ifdef __AVX__
        const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
        const %(idx_type)s* __restrict__ _idx = col_idx_.data();

        if (_transmission && %(post_prefix)s_active) {
            %(size_type)s _s, _stop;
            double _tmp_sum[4];
            double* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m256d _tmp_reg_sum = _mm256_setzero_pd();

                for (; _s+8 < _stop; _s+=8) {
                    __m256d _tmp_r = _mm256_set_pd(
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m256d _tmp_r2 = _mm256_set_pd(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]
                    );

                    _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _mm256_add_pd(_tmp_r, _tmp_r2));
                }

                _mm256_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
""",
        'float': """
    #ifdef __AVX__
        const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
        const %(idx_type)s* __restrict__ _idx = col_idx_.data();

        if (_transmission && %(post_prefix)s_active) {
            %(size_type)s _s, _stop;
            float _tmp_sum[8];
            float* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m256 _tmp_reg_sum = _mm256_setzero_ps();

                for (; _s+16 < _stop; _s+=16) {
                    __m256 _tmp_r = _mm256_set_ps(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m256 _tmp_r2 = _mm256_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]]
                    );

                    _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_add_ps(_tmp_r, _tmp_r2));
                }
                _mm256_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
    #endif
"""
    }
}

continuous_transmission_avx512_single_weight = {
    'sum' : {
        'double': """
    #ifdef __AVX__
        const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
        const %(idx_type)s* __restrict__ _idx = col_idx_.data();

        if (_transmission && %(post_prefix)s_active) {
            %(size_type)s _s, _stop;
            double _tmp_sum[8];
            double* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m512d _tmp_reg_sum = _mm512_setzero_pd();

                for (; (_s+8) < _stop; _s+=8) {
                    __m512d _tmp_r = _mm512_set_pd(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );

                    _tmp_reg_sum = _mm512_add_pd(_tmp_reg_sum, _tmp_r);
                }

                _mm512_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
    #endif
""",
        'float': """
    #ifdef __AVX__
        const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
        const %(idx_type)s* __restrict__ _idx = col_idx_.data();

        if (_transmission && %(post_prefix)s_active) {
            %(size_type)s _s, _stop;
            float _tmp_sum[16];
            float* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m512 _tmp_reg_sum = _mm512_setzero_ps();

                for (; (_s+16) < _stop; _s+=16) {
                    __m512 _tmp_r = _mm512_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]],
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );

                    _tmp_reg_sum = _mm512_add_ps(_tmp_reg_sum, _tmp_r);
                }
                _mm512_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7] + _tmp_sum[8] + _tmp_sum[9] + _tmp_sum[10] + _tmp_sum[11] + _tmp_sum[12] + _tmp_sum[13] + _tmp_sum[14] + _tmp_sum[15];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]];

                %(post_prefix)s_sum_%(target)s%(post_index)s += w * lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
"""
    }
}

###############################################################################
# Optimized kernel for default rate-coded continuous transmission using 
# SIMD instructions (SSE4_1, AVX, AVX-512s)
###############################################################################
continuous_transmission_sse = {
    'sum' : {
        'double': """
    #ifdef __SSE4_1__

        if (_transmission && %(post_prefix)s_active) {
            const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
            const %(idx_type)s* __restrict__ _idx = col_idx_.data();
            const double* __restrict__ _w = w.data();

            double _tmp_sum[2];
            double* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];
                %(size_type)s _s = row_ptr[rk_post];
                %(size_type)s _stop = row_ptr[rk_post+1];
                __m128d _tmp_reg_sum = _mm_setzero_pd();

                for (; _s+8 < _stop; _s+=8) {
                    __m128d _tmp_r = _mm_set_pd(_pre_r[_idx[_s+1]], _pre_r[_idx[_s+0]]);
                    __m128d _tmp_r2 = _mm_set_pd(_pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]]);
                    __m128d _tmp_r3 = _mm_set_pd(_pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]);
                    __m128d _tmp_r4 = _mm_set_pd(_pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]]);

                    __m128d _tmp_w = _mm_loadu_pd(&_w[_s]);
                    __m128d _tmp_w2 = _mm_loadu_pd(&_w[_s+2]);
                    __m128d _tmp_w3 = _mm_loadu_pd(&_w[_s+4]);
                    __m128d _tmp_w4 = _mm_loadu_pd(&_w[_s+6]);

                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r, _tmp_w));
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r2, _tmp_w2));
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r3, _tmp_w3));
                    _tmp_reg_sum = _mm_add_pd(_tmp_reg_sum, _mm_mul_pd(_tmp_r4, _tmp_w4));
                }

                _mm_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                %(post_prefix)s_sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
""",
        'float': """
    #ifdef __SSE4_1__
        if (_transmission && %(post_prefix)s_active) {
            const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
            const %(idx_type)s* __restrict__ _idx = col_idx_.data();
            const float* __restrict__ _w = w.data();

            float _tmp_sum[4];
            float* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];
                %(size_type)s _s = row_ptr[rk_post];
                %(size_type)s _stop = row_ptr[rk_post+1];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m128 _tmp_reg_sum = _mm_setzero_ps();

                for (; _s+16 < _stop; _s+=16) {
                    __m128 _tmp_r = _mm_set_ps(_pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]);
                    __m128 _tmp_r2 = _mm_set_ps(_pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]);
                    __m128 _tmp_r3 = _mm_set_ps(_pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]]);
                    __m128 _tmp_r4 = _mm_set_ps(_pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]]);

                    __m128 _tmp_w = _mm_loadu_ps(&_w[_s]);
                    __m128 _tmp_w2 = _mm_loadu_ps(&_w[_s+4]);
                    __m128 _tmp_w3 = _mm_loadu_ps(&_w[_s+8]);
                    __m128 _tmp_w4 = _mm_loadu_ps(&_w[_s+12]);

                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r, _tmp_w));
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r2, _tmp_w2));
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r3, _tmp_w3));
                    _tmp_reg_sum = _mm_add_ps(_tmp_reg_sum, _mm_mul_ps(_tmp_r4, _tmp_w4));
                }
                _mm_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                %(post_prefix)s_sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with SSE4-1 support. Please check your compiler flags ..." << std::endl;
    #endif
"""
    }
}

continuous_transmission_avx = {
    'sum' : {
        'double': """
    #ifdef __AVX__
        const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
        const %(idx_type)s* __restrict__ _idx = col_idx_.data();
        const double* __restrict__ _w = w.data();

        if (_transmission && %(post_prefix)s_active) {
            %(size_type)s _s, _stop;
            double _tmp_sum[4];
            double* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m256d _tmp_reg_sum = _mm256_setzero_pd();

                for (; _s+8 < _stop; _s+=8) {
                    __m256d _tmp_r = _mm256_set_pd(
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m256d _tmp_r2 = _mm256_set_pd(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]]
                    );

                    __m256d _tmp_w = _mm256_loadu_pd(&_w[_s]);
                    __m256d _tmp_w2 = _mm256_loadu_pd(&_w[_s+4]);

                    _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _mm256_mul_pd(_tmp_r, _tmp_w));
                    _tmp_reg_sum = _mm256_add_pd(_tmp_reg_sum, _mm256_mul_pd(_tmp_r2, _tmp_w2));
                }

                _mm256_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                %(post_prefix)s_sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
""",
        'float': """
    #ifdef __AVX__
        const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
        const %(idx_type)s* __restrict__ _idx = col_idx_.data();
        const float* __restrict__ _w = w.data();

        if (_transmission && %(post_prefix)s_active) {
            %(size_type)s _s, _stop;
            float _tmp_sum[8];
            float* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m256 _tmp_reg_sum = _mm256_setzero_ps();

                for (; _s+16 < _stop; _s+=16) {
                    __m256 _tmp_r = _mm256_set_ps(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );
                    __m256 _tmp_r2 = _mm256_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]]
                    );

                    __m256 _tmp_w = _mm256_loadu_ps(&_w[_s]);
                    __m256 _tmp_w2 = _mm256_loadu_ps(&_w[_s+8]);

                    _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_mul_ps(_tmp_r, _tmp_w));
                    _tmp_reg_sum = _mm256_add_ps(_tmp_reg_sum, _mm256_mul_ps(_tmp_r2, _tmp_w2));
                }
                _mm256_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                %(post_prefix)s_sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX support. Please check your compiler flags ..." << std::endl;
    #endif
"""
    }
}

continuous_transmission_avx512 = {
    'sum' : {
        'double': """
    #ifdef __AVX512F__
        const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
        const %(idx_type)s* __restrict__ _idx = col_idx_.data();
        const double* __restrict__ _w = w.data();

        if (_transmission && %(post_prefix)s_active) {
            %(size_type)s _s, _stop;
            double _tmp_sum[8];
            double* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m512d _tmp_reg_sum = _mm512_setzero_pd();

                for (; (_s+8) < _stop; _s+=8) {
                    __m512d _tmp_r = _mm512_set_pd(
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );

                    __m512d _tmp_w = _mm512_loadu_pd(&_w[_s]);

                    _tmp_reg_sum = _mm512_add_pd(_tmp_reg_sum, _mm512_mul_pd(_tmp_r, _tmp_w));
                }

                _mm512_storeu_pd(_tmp_sum, _tmp_reg_sum);

                // partial sums
                double lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                %(post_prefix)s_sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
    #endif
""",
        'float': """
    #ifdef __AVX512F__
        const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
        const %(idx_type)s* __restrict__ _idx = col_idx_.data();
        const float* __restrict__ _w = w.data();

        if (_transmission && %(post_prefix)s_active) {
            %(size_type)s _s, _stop;
            float _tmp_sum[16];
            float* __restrict__ _pre_r = %(get_r)s;

            for (%(idx_type)s i = 0; i < post_ranks_.size(); i++) {
                %(idx_type)s rk_post = post_ranks_[i];

                _s = row_ptr[rk_post];
                _stop = row_ptr[rk_post+1];
                __m512 _tmp_reg_sum = _mm512_setzero_ps();

                for (; (_s+16) < _stop; _s+=16) {
                    __m512 _tmp_r = _mm512_set_ps(
                        _pre_r[_idx[_s+15]], _pre_r[_idx[_s+14]], _pre_r[_idx[_s+13]], _pre_r[_idx[_s+12]],
                        _pre_r[_idx[_s+11]], _pre_r[_idx[_s+10]], _pre_r[_idx[_s+9]], _pre_r[_idx[_s+8]],
                        _pre_r[_idx[_s+7]], _pre_r[_idx[_s+6]], _pre_r[_idx[_s+5]], _pre_r[_idx[_s+4]],
                        _pre_r[_idx[_s+3]], _pre_r[_idx[_s+2]], _pre_r[_idx[_s+1]], _pre_r[_idx[_s]]
                    );

                    __m512 _tmp_w = _mm512_loadu_ps(&_w[_s]);

                    _tmp_reg_sum = _mm512_add_ps(_tmp_reg_sum, _mm512_mul_ps(_tmp_r, _tmp_w));
                }
                _mm512_storeu_ps(_tmp_sum, _tmp_reg_sum);

                // partial sums
                float lsum = _tmp_sum[0] + _tmp_sum[1] + _tmp_sum[2] + _tmp_sum[3] + _tmp_sum[4] + _tmp_sum[5] + _tmp_sum[6] + _tmp_sum[7] + _tmp_sum[8] + _tmp_sum[9] + _tmp_sum[10] + _tmp_sum[11] + _tmp_sum[12] + _tmp_sum[13] + _tmp_sum[14] + _tmp_sum[15];

                // remainder loop
                for (; _s < _stop; _s++)
                    lsum += _pre_r[_idx[_s]] * _w[_s];

                %(post_prefix)s_sum_%(target)s%(post_index)s += lsum;
            }
        } // active
    #else
        std::cerr << "The code was not compiled with AVX-512 support. Please check your compiler flags ..." << std::endl;
    #endif
"""
    }
}

###############################################################################
# Continuous synaptic update
###############################################################################
update_variables = {
    'local': """
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L) ){
    // global variables
    %(global)s

    const %(size_type)s* __restrict__ row_ptr = row_begin_.data();
    const %(idx_type)s* __restrict__ col_idx = col_idx_.data();

    %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_ranks_.size());
    for (%(idx_type)s i = 0; i < nb_post; i++) {
        rk_post = post_ranks_[i];

        // semiglobal variables
    %(semiglobal)s
    
        // local variables
        for(%(size_type)s j = row_ptr[rk_post]; j < row_ptr[rk_post+1]; j++){
            rk_pre = col_idx[j];
    %(local)s
        }
    }
}
""",
    'global': """
if(_transmission && _update && %(post_prefix)s_active && ( (t - _update_offset)%%_update_period == 0L)){
    %(global)s
    
    %(idx_type)s nb_post = static_cast<%(idx_type)s>(post_ranks_.size());
    for (%(idx_type)s i = 0; i < nb_post; i++) {
        %(idx_type)s rk_post = post_ranks_[i];

    %(semiglobal)s
    }
}
"""
}

###############################################################
# Event-driven updates
###############################################################
spiking_summation_fixed_delay_csr = """// Event-based summation
if (_transmission && %(post_prefix)s_active) {
    // w as CSR
    const int * __restrict__ col_ptr = _col_ptr.data();

    // Iterate over all spiking neurons
    for (int _idx = 0; _idx < %(pre_array)s.size(); _idx++) {
        int _pre = %(pre_array)s[_idx];

        // slice in CSRC
        int beg = col_ptr[_pre];
        int end = col_ptr[_pre+1];

        // Iterate over connected post neurons
        for (int syn = beg; syn < end; syn++) {
            %(event_driven)s
            %(g_target)s
            %(pre_event)s
        }
    }
} // active
"""

spiking_post_event = """
// w as CSR
const int * __restrict__ row_ptr = row_begin_.data();
const int * __restrict__ col_idx = col_idx_.data();

if(_transmission && %(post_prefix)s_active){
    int rk_post, beg, end;

    for(int _idx_i = 0; _idx_i < %(post_prefix)sspiked.size(); _idx_i++){
        // Rank of the postsynaptic neuron which fired
        rk_post = %(post_prefix)sspiked[_idx_i];

        // slice in CSRC
        beg = row_ptr[rk_post];
        end = row_ptr[rk_post+1];

        // Iterate over all synapse to this neuron
        for (int j = beg; j < end; j++) {
            int rk_pre = col_idx[j];

%(event_driven)s
%(post_event)s
        }
    }
}
"""

conn_templates = {
    # accessors
    'attribute_decl': attribute_decl,
    'attribute_cpp_init': attribute_cpp_init,
    'attribute_cpp_size': attribute_cpp_size,
    'attribute_cpp_delete': attribute_cpp_delete,
    'delay': delay,
    'event_driven': event_driven,

    # operations
    'rate_coded_sum': csr_summation_operation,
    'vectorized_default_psp': {
        'sse': {
            'single_w': continuous_transmission_sse_single_weight,
            'multi_w': continuous_transmission_sse
        },
        'avx': {
            'single_w': continuous_transmission_avx_single_weight,
            'multi_w': continuous_transmission_avx
        },
        'avx512': {
            'single_w': continuous_transmission_avx512_single_weight,
            'multi_w': continuous_transmission_avx512
        }
    },
    'update_variables': update_variables,
    'spiking_sum_fixed_delay': spiking_summation_fixed_delay_csr,
    'spiking_sum_variable_delay': None,
    'post_event': spiking_post_event
}

conn_ids = {
    'local_index': '[j]',
    'semiglobal_index': '[i]',
    'global_index': '',
    'pre_index': '[rk_pre]',
    'post_index': '[rk_post]',
    'delay_nu' : '[delay[j]-1]', # non-uniform delay
    'delay_u' : '[delay-1]' # uniform delay
}


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Constant import Constant
from ANNarchy.intern.NetworkManager import NetworkManager
from ANNarchy.intern.Profiler import Profiler
from ANNarchy.intern.ConfigManagement import ConfigManager, _check_paradigm
from ANNarchy.intern import Messages

from ANNarchy.core.PopulationView import PopulationView
from ANNarchy.core.Random import *
from ANNarchy.core.Neuron import Neuron, IndividualNeuron

from typing import Iterator
import numpy as np
import copy, inspect
import re


class Population :
    """
    Population of neurons.
    
    The object is returned by `Network.create()` and should not be created directly.

    ```python
    net = ann.Network()
    pop = net.create(100, neuron=ann.Izhikevich, name="Excitatory population")
    ```
    
    :param geometry: population geometry as tuple. If an integer is given, it is the size of the population.
    :param neuron: `Neuron` instance. It can be user-defined or a built-in model.
    :param name: unique name of the population (optional).
    :param stop_condition: a single condition on a neural variable which can stop the simulation whenever it is true.
    """

    def __init__(
            self, 
            geometry: tuple | int, 
            neuron: "Neuron", 
            name:str = None, 
            stop_condition:str = None, 
            # Internal use only
            storage_order:str = 'post_to_pre', 
            copied:bool = False,
            net_id:int = 0,
        ):

        # Check if the network has already been compiled
        if NetworkManager().get_network(net_id).compiled and not copied:
            Messages._error('You cannot add a population after the network has been compiled.')

        # Store the provided geometry. Automatically defines w, h, d, size
        self.geometry = geometry
        "Geometry of the population."
        self.width = 0
        "Width of the population."
        self.height = 0
        "Height of the population."
        self.depth = 0
        "Depth of the population."
        self.dimension = 0
        "Number of dimensions of the population."

        if isinstance(geometry, (int, float)):
            # 1D
            self.geometry = (int(geometry), )
            self.width = int(geometry)
            self.height = int(1)
            self.depth = int(1)
            self.dimension = int(1)

        elif isinstance(geometry, tuple):
            # a tuple is given, can be 1 .. N dimensional
            self.geometry = ()
            for d in geometry:
                self.geometry += (int(d),)
            self.width = int(geometry[0])
            if len(geometry)>=2:
                self.height = int(geometry[1])
            else:
                self.height = int(1)
            if len(geometry)>=3:
                self.depth = int(geometry[2])
            else:
                self.depth = int(1)

            self.dimension = len(geometry)
        else:
            Messages._error('Population(): the geometry must be either an integer or a tuple.')

        # Compute the size
        size = int(1)
        for i in range(len(self.geometry)):
            size *= int(self.geometry[i])
        self.size = int(size)
        "Size of the population (total number of neurons)."
        self.ranks = np.arange(self.size, dtype="int32")
        "Array of ranks in the population (between 0 and `size - 1`)."

        # Network
        self.net_id = net_id

        # Store the neuron type
        if inspect.isclass(neuron):
            self.neuron_type = neuron()
        else:
            self.neuron_type = copy.deepcopy(neuron)
        self.neuron_type._analyse(self.net_id)

        # Store the stop condition
        self.stop_condition = stop_condition


        # Attribute a name if not provided
        self.id = NetworkManager().get_network(net_id)._add_population(self)
        self.class_name = 'pop'+str(self.id)

        if name:
            if re.match(r'^pop\d+$', name):
                # Name is already pop0, pop1...
                self.name = self.class_name
            else:
                self.name = name
                "Name of the population"
        else:
            self.name = self.class_name

        # Get a list of parameters and variables
        self.parameters = []
        "List of parameter names."
        self.variables = []
        "List of variable names."
        for param in self.neuron_type.description['parameters']:
            self.parameters.append(param['name'])
        for var in self.neuron_type.description['variables']:
            self.variables.append(var['name'])
        self.attributes = self.parameters + self.variables
        "List of attribute names (parameters + variables)."

        # Get a list of user-defined functions
        self.functions = [func['name'] for func in self.neuron_type.description['functions']]
        "List of functions defined by the neuron model."

        # Store initial values
        self.init = {}
        for param in self.neuron_type.description['parameters']:
            self.init[param['name']] = param['init']
        for var in self.neuron_type.description['variables']:
            self.init[var['name']] = var['init']

        # List of targets actually connected
        self.targets = []
        "List of connected targets."

        # List of global operations needed by connected projections
        self.global_operations = []

        # Maximum delay of connected projections
        self.max_delay = 0

        # Spiking neurons: do they have to compute an average?
        self._compute_mean_fr = -1.

        # Finalize initialization
        self.initialized = False
        self.cyInstance = None
        self.enabled = True

        # Rank <-> Coordinates methods
        # for the one till three dimensional case we use cython optimized functions.
        from ANNarchy.cython_ext import Coordinates
        if self.dimension==1:
            self._rank_from_coord = Coordinates.get_rank_from_1d_coord
            self._coord_from_rank = Coordinates.get_1d_coord
        elif self.dimension==2:
            self._rank_from_coord = Coordinates.get_rank_from_2d_coord
            self._coord_from_rank = Coordinates.get_2d_coord
        elif self.dimension==3:
            self._rank_from_coord = Coordinates.get_rank_from_3d_coord
            self._coord_from_rank = Coordinates.get_3d_coord
        else:
            self._rank_from_coord = np.ravel_multi_index
            self._coord_from_rank = np.unravel_index

        self._norm_coord_dict = {
            1: Coordinates.get_normalized_1d_coord,
            2: Coordinates.get_normalized_2d_coord,
            3: Coordinates.get_normalized_3d_coord
        }

        # Recorded variables
        self._monitor = None

        # Is overwritten by SpecificPopulations
        self._specific_template = {}

        # Storage order. 
        self._storage_order = storage_order

    def _copy(self, net_id=None):
        "Returns a copy of the population when creating networks. Internal use only."
        return Population(
            geometry=self.geometry, 
            neuron=self.neuron_type, 
            name=self.name, 
            stop_condition=self.stop_condition, 
            storage_order=self._storage_order, 
            copied=True,
            net_id = self.net_id if net_id is None else net_id,
        )

    def _generate(self):
        "Overriden by specific populations to generate the code."
        pass

    def _instantiate(self, module):
        """
        Instantiates the population after compilation of the C++ simulation core.
        The function should solely called by Compiler._instantiate().

        :param module: cython module (ANNarchyCore instance)
        """
        if Profiler().enabled:
            import time
            t1 = time.time()

        try:
            self.cyInstance = getattr(module, self.class_name+'_wrapper')(self.size, self.max_delay)
        except:
            Messages._error('unable to instantiate the population', self.name)

        if Profiler().enabled:
            t2 = time.time()
            Profiler().add_entry(t1, t2, "pop"+str(self.id), "instantiate")

    def _init_attributes(self):
        """ Method used after compilation to initialize the attributes."""
        
        # Initialize the population
        self.initialized = True

        # Transfer the initial values of all attributes
        for name, value in self.init.items():
            if isinstance(value, Constant):
                self.__setattr__(name, value.value)
            elif isinstance(value, RandomDistribution): # The initial value of a variable is a random variable
                self.__setattr__(name, value.get_list_values(self.size))
            elif isinstance(value, str): # The initial value of a variable is a parameter
                self.__setattr__(name, self.__getattr__(value))
            else:
                self.__setattr__(name, value)

        # Activate the population
        self.cyInstance.activate(self.enabled)

        # Reset to generate the right structures
        self.cyInstance.reset()

        # If the spike population has a refractory period:
        if self.neuron_type.type == 'spike' and self.neuron_type.description['refractory']:
            if not isinstance(self.neuron_type.description['refractory'], str): # the variable will be used directly
                self.refractory = self.neuron_type.description['refractory']

        # Spiking neurons can compute a mean FR
        if self.neuron_type.type == 'spike':
            getattr(self.cyInstance, 'compute_firing_rate')(self._compute_mean_fr)

    def _size_in_bytes(self) -> int:
        """
        Returns the size of allocated memory on the C++ side. Please note that this does not contain monitored data and works only if compile() has been invoked.
        """
        if self.initialized:
            return self.cyInstance.size_in_bytes()
        else:
            return 0

    def _clear(self):
        """
        Deallocates container within the C++ instance. The population object is not usable anymore after calling this function.

        Warning: should be only called by the net deconstruction ( in context of parallel_run() ).
        """
        if self.initialized:
            self.cyInstance.clear()
            self.initialized = False

    def reset(self, attributes:list = None)  -> None:
        """
        Resets all parameters and variables of the population to the value they had before the call to `net.compile()`.

        :param attributes: list of attributes (parameter or variable) which should be reinitialized. Default: all attributes.
        """
        if attributes is None:
            try:
                self.set(self.init)
            except Exception as e:
                Messages._print(e)
                Messages._error("Population.reset(): something went wrong while resetting.")
        else: # only some of them
            for var in attributes:
                # check it exists
                if not var in self.attributes:
                    Messages._warning("Population.reset():", var, "is not an attribute of the population, skipping.")
                    continue

                try:
                    self.__setattr__(var, self.init[var])
                except Exception as e:
                    Messages._print(e)
                    Messages._warning("Population.reset(): something went wrong while resetting", var)

        self.cyInstance.activate(self.enabled)
        self.cyInstance.reset()

    def clear(self) -> None:
        """
        Clears all spiking events previously emitted (history of spikes, delayed spikes).

        Can be useful if you do not want to totally reset a population (i.e. all variables), only to clear the spiking history between two trials.

        Note: does nothing for rate-coded networks.
        """
        self.cyInstance.reset()

    def enable(self) -> None:
        """
        (Re)-enables computations in this population, after they were disabled by the `disable()` method.

        The status of the population is accessible through the `enabled` flag.
        """
        if self.initialized:
            self.cyInstance.activate(True)
        self.enabled = True

    def disable(self) -> None:
        """
        Temporarily disables computations in this population (including the projections leading to it).

        You can re-enable it with the `enable()` method.
        """
        if self.initialized:
            self.cyInstance.activate(False)
        self.enabled = False

    def __getattr__(self, name):
        # Method called when accessing an attribute.
        if name == 'initialized' or not hasattr(self, 'initialized'): # Before the end of the constructor
            return object.__getattribute__(self, name)
        elif name == 'spike':
            if not self.initialized:
                Messages._error("Accessing spike events is only valid after compile()")
            else:
                return self._get_cython_attribute(name)
        elif hasattr(self, 'attributes'):
            if name in self.attributes:
                if self.initialized: # access after compile()
                    return self._get_cython_attribute(name)
                else: # access before compile()
                    if name in self.neuron_type.description['local']:
                        if isinstance(self.init[name], np.ndarray):
                            return self.init[name]
                        else:
                            return np.array([self.init[name]] * self.size).reshape(self.geometry)
                    else:
                        return self.init[name]
            elif name in self.functions:
                return self._function(name)
            else:
                return object.__getattribute__(self, name)
        return object.__getattribute__(self, name)

    def __setattr__(self, name, value):
        # Method called when setting an attribute.
        if name == 'initialized' or not hasattr(self, 'initialized'): # Before the end of the constructor
            object.__setattr__(self, name, value)
        elif hasattr(self, 'attributes'):
            if name in self.attributes:
                if not self.initialized:
                    if isinstance(value, RandomDistribution): # Make sure it is generated only once
                        self.init[name] = np.array(value.get_values(self.size)).reshape(self.geometry)
                    else:
                        self.init[name] = value
                else:
                    self._set_cython_attribute(name, value)
            else:
                object.__setattr__(self, name, value)
        else:
            object.__setattr__(self, name, value)

    def _get_cython_attribute(self, attribute):
        """
        Returns the value of the given attribute for all neurons in the population,
        as a Numpy array having the same geometry as the population if it is local.

        :param attribute: should be a string representing the variables's name.
        """
        try:
            if attribute in self.neuron_type.description['local']:
                data = np.array(getattr(self.cyInstance, attribute))
                return data.reshape(self.geometry)
            else:
                return getattr(self.cyInstance, attribute)
        except Exception as e:
            Messages._print(e)
            Messages._error(' the variable ' +  attribute +  ' does not exist in this population (' + self.name + ')')

    def _set_cython_attribute(self, attribute, value):
        """
        Sets the value of the given attribute for all neurons in the population,
        as a Numpy array having the same geometry as the population if it is local.

        :param attribute: should be a string representing the variables's name.
        :param value: a value or Numpy array of the right size.
        """

        try:
            ctype = self._get_attribute_cpp_type(attribute)
            if attribute in self.neuron_type.description['local']:
                if isinstance(value, np.ndarray):
                    if _check_paradigm("cuda", self.net_id):
                        if ctype == "char":
                            raise NotImplementedError
                        setattr(self.cyInstance, attribute, value.reshape(self.size).tolist())
                        setattr(self.cyInstance, attribute+"_host_to_device", True)
                    else:
                        setattr(self.cyInstance, attribute, value.reshape(self.size).tolist())

                elif isinstance(value, list):
                    
                    if _check_paradigm("cuda", self.net_id):
                        # See `ANNarchy.parser.Extraction.extract_boundsflags` for more details
                        if ctype == "char":
                            value = [ str(0) if x == False else str(1) for x in value ]
                        setattr(self.cyInstance, attribute, value)
                        setattr(self.cyInstance, attribute+"_host_to_device", True)
                    else:
                        setattr(self.cyInstance, attribute, value)

                else:
                    if _check_paradigm("cuda", self.net_id):
                        # See `ANNarchy.parser.Extraction.extract_boundsflags` for more details
                        if ctype == "char":
                            value = str(0) if value == False else str(1)
                        setattr(self.cyInstance, attribute, [value for _ in range(self.size) ])                        
                        setattr(self.cyInstance, attribute+"_host_to_device", True)
                    else:
                        setattr(self.cyInstance, attribute, [value for _ in range(self.size) ])

            else:
                setattr(self.cyInstance, attribute, value)
                if _check_paradigm("cuda", self.net_id):
                    # Variables need to be updated during next simulate() call.
                    if attribute in [attr['name'] for attr in self.neuron_type.description['variables']]:
                        setattr(self.cyInstance, attribute+"_host_to_device", True)

        except Exception as e:
            Messages._print(e)
            err_msg = """Population.set(): either the variable '%(attr)s' does not exist in the population '%(pop)s', or the provided array does not have the right size."""
            Messages._error(err_msg  % { 'attr': attribute, 'pop': self.name } )

    def _get_attribute_cpp_type(self, attribute):
        """
        Determines C++ data type for a given attribute
        """
        ctype = None
        for var in self.neuron_type.description['variables']+self.neuron_type.description['parameters']:
            if var['name'] == attribute:
                ctype = var['ctype']
                break

        return ctype

    def __len__(self):
        # Number of neurons in the population.
        return self.size


    def set(self, values:dict) -> None:
        """
        Sets the value of neural variables and parameters from a dictionary.

        Example:

        ```python
        pop.set({'tau': 20.0, 'r': np.random.rand((8,8)) } )
        ```

        :param values: dictionary of attributes to be updated.
        """
        for name, value in values.items():
            self.__setattr__(name, value)

    def get(self, name:str) -> np.ndarray:
        """
        Returns the value of a neural variable or parameter based on its name.

        :param name: attribute name as a string.
        """
        return self.__getattr__(name)



    ################################
    ## Access to functions
    ################################
    def _function(self, name):
        "Access a user defined function"
        if not self.initialized:
            Messages._warning('the network is not compiled yet, cannot access the function ' + name)
            return
        
        # Get the C++ function
        cpp_function = getattr(self.cyInstance, name)

        # One argument
        def apply(*args):
            return list(map(cpp_function, *args))
        return apply

    ################################
    ## Access to weighted sums
    ################################
    def sum(self, target:str) -> np.ndarray:
        """
        Returns the array of weighted sums corresponding to the target:

        ```python
        excitatory = pop.sum('exc')
        ```

        For spiking networks, this is equivalent to accessing the conductances directly:

        ```python
        excitatory = pop.g_exc
        ```

        If no incoming projection has the given target, the method returns zeros.

        **Note:** it is not possible to distinguish the original population when the same target is used.

        :param target: the desired projection target.
        """
        # Check if the network is initialized
        if not self.initialized:
            Messages._warning('sum(): the population', self.name, 'is not initialized yet.')
            return np.zeros(self.geometry)
        
        # Check if a projection has this type
        if not target in self.targets:
            Messages._warning('sum(): the population', self.name, 'receives no projection with the target', target)
            return np.zeros(self.geometry)
        
        # Spiking neurons already have conductances available
        if self.neuron_type.type == 'spike':
            return np.array(getattr(self, 'g_'+target))
        
        # Otherwise, call the Cython method
        return np.array(getattr(self.cyInstance, "_sum_"+target))

    ################################
    ## Refractory period
    ################################
    @property
    def refractory(self) -> float | str:
        "Refractory period (in ms)."
        if self.neuron_type.description['type'] == 'spike':
            if self.initialized:
                if not isinstance(self.neuron_type.description['refractory'], str):
                    return ConfigManager().get('dt', self.net_id)*self.cyInstance.get_refractory()
                else:
                    return getattr(self, self.neuron_type.description['refractory'])
            else :
                return self.neuron_type.description['refractory']
        else:
            Messages._warning('Rate-coded neurons do not have refractory periods...')
            return None

    @refractory.setter
    def refractory(self, value:float | str):
        if self.neuron_type.description['type'] == 'spike':

            if isinstance(self.neuron_type.description['refractory'], str):
                Messages._warning("The refractory period is linked to the neural variable", self.neuron_type.description['refractory'], ", doing nothing... Change its value instead.")
                return

            if self.initialized:
                if isinstance(value, RandomDistribution):
                    refs = (value.get_values(self.size)/ConfigManager().get('dt', self.net_id)).astype(int)
                elif isinstance(value, np.ndarray):
                    refs = (value / ConfigManager().get('dt', self.net_id)).astype(int).reshape(self.size)
                else:
                    refs = (value/ ConfigManager().get('dt', self.net_id)*np.ones(self.size)).astype(int)

                self.cyInstance.refractory = refs
                if _check_paradigm("cuda", self.net_id):
                    self.cyInstance.refractory_dirty = True

            else: # not initialized yet, saving for later
                self.neuron_type.description['refractory'] = value
        else:
            Messages._warning('Rate-coded neurons do not have refractory periods...')

    ################################
    ## Spiking neurons can compute a mean FR
    ################################
    def compute_firing_rate(self, window:float) -> None:
        """
        Tells spiking neurons in the population to compute their mean firing rate over the given window and store the values in the variable `r`.

        This method has an effect on spiking neurons only.

        If this method is not called, `r` will always be 0.0. `r` can of course be accessed and recorded as any other variable.

        :param window: window in ms over which the spikes will be counted.
        """
        if _check_paradigm('cuda', self.net_id):
            Messages._warning('compute_firing_rate() is currently being evaluated on the host-side, so may be slow ... ')

        if self.neuron_type.type == 'rate':
            Messages._error('compute_firing_rate(): the neuron is already rate-coded...')

        self._compute_mean_fr = float(window)

        if self.initialized:
            getattr(self.cyInstance, 'compute_firing_rate')(self._compute_mean_fr)

    ################################
    ## Access to individual neurons
    ################################
    def neuron(self, *coord) -> IndividualNeuron:
        """
        Returns an ``IndividualNeuron`` object wrapping the neuron with the provided rank or coordinates.
        """
        # Transform arguments
        if len(coord) == 1:
            if isinstance(coord[0], int):
                rank = coord[0]
                if not rank < self.size:
                    Messages._error(' when accessing neuron', str(rank), ': the population', self.name, 'has only', self.size, 'neurons (geometry '+ str(self.geometry) +').')
            else:
                rank = self.rank_from_coordinates( coord[0] )
                if rank is None:
                    return None
        else: # a tuple
            rank = self.rank_from_coordinates( coord )
            if rank is None:
                return None

        # Return corresponding neuron
        return IndividualNeuron(self, rank)

    @property
    def neurons(self) -> Iterator[IndividualNeuron]:
        """
        Returns iteratively each neuron in the population.

        For instance, if you want to iterate over all neurons of a population:

        ```python
        for neuron in pop.neurons:
            neuron.r = 0.0
        ```

        Alternatively, one could also benefit from the ``__iter__`` special command. 
        The following code is equivalent:

        ```python
        for neuron in pop:
            neuron.r = 0.0
        ```
        """
        for neur_rank in range(self.size):
            yield self.neuron(neur_rank)

    # Iterators
    def __getitem__(self, *args, **kwds):
        """ 
        Returns neurons froms coordinates in the population.

        If only one argument is given, it is interpeted as a rank and returns a single neuron.

        If slices are given, it returns a PopulationView object.
        """
        indices =  args[0]
        try:
            if np.issubdtype(indices, int):
                indices = int(indices)
        except:
            pass

        if isinstance(indices, int): # a single neuron
            return PopulationView(self, ranks=np.array([int(indices)]), geometry=(1,))

        elif isinstance(indices, (list)):
            indices = np.array(indices)
            return PopulationView(self, indices, geometry=(len(indices),))

        elif isinstance(indices, (np.ndarray)):
            # Sanity check
            if isinstance(indices, (np.ndarray)):
                if indices.ndim != 1:
                    Messages._error('only one-dimensional lists/arrays are allowed to address a population.')

            return PopulationView(self, indices, geometry=(len(indices),))

        elif isinstance(indices, slice): # a single slice of ranks
            start, stop, step = indices.start, indices.stop, indices.step

            # no value defined for a position
            if indices.start is None:
                start = 0
            if indices.stop is None:
                stop = self.size
            if indices.step is None:
                step = 1

            # start or end arguments are negative
            if start < 0:
                start = self.size + start
            if stop < 0:
                stop = self.size + stop

            # generate a new set of indices
            rk_range = np.arange(start, stop, step, dtype="int32")
            return PopulationView(self, rk_range, geometry=(len(rk_range),))

        elif isinstance(indices, tuple): # a tuple
            slices = False
            for idx in indices: # check if there are slices in the coordinates
                if isinstance(idx, slice): # there is at least one
                    slices = True
            if not slices: # return one neuron
                return self.neuron(indices)
            else: # Compute a list of ranks from the slices
                coords = []
                # Expand the slices
                for rank in range(len(indices)):
                    idx = indices[rank]
                    if isinstance(idx, int): # no slice
                        coords.append([idx])
                    elif isinstance(idx, slice): # slice
                        start, stop, step = idx.start, idx.stop, idx.step
                        if idx.start is None:
                            start = 0
                        if idx.stop is None:
                            stop = self.geometry[rank]
                        if idx.step is None:
                            step = 1
                        rk_range = list(range(start, stop, step))
                        coords.append(rk_range)
                # Generate all ranks from the indices
                if self.dimension == 2:
                    ranks = [self.rank_from_coordinates((x, y)) for x in coords[0] for y in coords[1]]
                    geometry = (len(coords[0]), len(coords[1]))
                elif self.dimension == 3:
                    ranks = [self.rank_from_coordinates((x, y, z)) for x in coords[0] for y in coords[1] for z in coords[2]]
                    geometry = (len(coords[0]), len(coords[1]), len(coords[2]))
                elif self.dimension == 4:
                    ranks = [self.rank_from_coordinates((x, y, z, k)) for x in coords[0] for y in coords[1] for z in coords[2] for k in coords[3]]
                    geometry = (len(coords[0]), len(coords[1]), len(coords[2]), len(coords[3]))
                else:
                    Messages._error("Slicing is implemented only for population with 4 dimensions at maximum", self.geometry)
                if not max(ranks) < self.size:
                    Messages._error("Indices do not match the geometry of the population", self.geometry)
                
                return PopulationView(self, ranks, geometry=geometry)

        Messages._warning('Population' + self.name + ': can not address the population with', indices)
        return None

    def __iter__(self) -> Iterator[IndividualNeuron]:
        # Returns iteratively each neuron in the population in ascending rank order.
        for neur_rank in range(self.size):
            yield self.neuron(neur_rank)

    ################################
    ## Coordinate transformations
    ################################
    def rank_from_coordinates(self, coord:tuple) -> int:
        """
        Returns the rank of a neuron based on coordinates.

        :param coord: coordinate tuple, can be multidimensional.
        """
        try:
            rank = self._rank_from_coord( coord, self.geometry )
        except:
            Messages._error('rank_from_coordinates(): There is no neuron of coordinates', coord, 'in the population', self.name, self.geometry)

        if rank > self.size:
            Messages._error('rank_from_coordinates(), neuron', str(coord), ': the population' , self.name , 'has only', self.size, 'neurons (geometry '+ str(self.geometry) +').')
        else:
            return rank

    def coordinates_from_rank(self, rank:int) -> tuple:
        """
        Returns the coordinates of a neuron based on its rank.

        :param rank: rank of the neuron.
        """
        # Check the rank
        if not rank < self.size:
            Messages._error('The given rank', str(rank), 'is larger than the size of the population', str(self.size) + '.')

        try:
            coord = self._coord_from_rank( rank, self.geometry )
        except:
            Messages._error('The given rank', str(rank), 'is larger than the size of the population', str(self.size) + '.')
        else:
            return coord

    def normalized_coordinates_from_rank(self, rank:int, norm:float=1.) -> tuple:
        """
        Returns normalized coordinates of a neuron based on its rank. 
        
        The geometry of the population is mapped to the hypercube $[0, 1]^d$

        :param rank: rank of the neuron.
        :param norm: norm of the hypercube (default = 1.0).
        """
        try:
            normal = self._norm_coord_dict[self.dimension](rank, self.geometry)
        except KeyError:
            coord = self.coordinates_from_rank(rank)

            normal = tuple()
            for dim in range(self.dimension):
                if self._geometry[dim] > 1:
                    normal += ( norm * float(coord[dim])/float(self.geometry[dim]-1), )
                else:
                    normal += (float(rank)/(float(self.size)-1.0),) # default?

        return normal


    ################################
    ## Save/load methods
    ################################
    def _data(self):
        "Returns a dictionary containing all information about the population. Used for saving."
        desc = {}
        desc['name'] = self.name
        desc['geometry'] = self.geometry
        desc['size'] = self.size
        
        # Attributes
        desc['attributes'] = self.attributes
        desc['parameters'] = self.parameters
        desc['variables'] = self.variables
        
        # Save all attributes
        for var in self.neuron_type.description['parameters'] + self.neuron_type.description['variables']:
            try:
                if var['locality'] == 'local':

                    if var['ctype'] == 'bool':
                        dtype = np.dtype(bool)
                    elif var['ctype'] == 'int':
                        dtype = np.dtype(int)
                    else:
                        dtype = np.dtype(float)

                    data = np.array(getattr(self.cyInstance, var['name']), dtype=dtype)
                    desc[var['name']] = data.reshape(self.geometry)
                else:
                    desc[var['name']] = getattr(self.cyInstance, var['name'])

            except:
                Messages._warning('Can not save the attribute ' + var['name'] + ' in the population ' + self.name + '.')

        return desc

    def save(self, filename:str) -> None:
        """
        Saves all information about the population (structure, current value of parameters and variables) into a file.

        * If the file name is '.npz', the data will be saved and compressed using `np.savez_compressed` (recommended).
        * If the file name ends with '.gz', the data will be pickled into a binary file and compressed using gzip.
        * If the file name is '.mat', the data will be saved as a Matlab 7.2 file. Scipy must be installed.
        * Otherwise, the data will be pickled into a simple binary text file using pickle.

        **Warning:** The '.mat' data will not be loadable by ANNarchy, it is only for external analysis purpose.

        Example:

        ```python
        pop.save('pop1.npz')
        pop.save('pop1.txt')
        pop.save('pop1.txt.gz')
        pop.save('pop1.mat')
        ```

        :param filename: Filename, may contain relative or absolute path.
        """
        from ANNarchy.core.IO import _save_data
        _save_data(filename, self._data())


    def load(self, filename, pickle_encoding=None) -> None:
        """
        Load the saved state of the population by `Population.save()`.

        Warning: Matlab data can not be loaded.

        Example:

        ```python
        pop.load('pop1.npz')
        pop.load('pop1.txt')
        pop.load('pop1.txt.gz')
        ```

        :param filename: the filename with relative or absolute path.
        """
        from ANNarchy.core.IO import _load_data
        self._load_pop_data(_load_data(filename, pickle_encoding))

    def _load_pop_data(self, desc):
        """
        Updates the population with the stored data set.
        """
        if not 'attributes' in desc.keys():
            Messages._error('Saved with a too old version of ANNarchy (< 4.2).', exit=True)

        for var in desc['attributes']:
            try:
                if isinstance(desc[var], (np.ndarray)):
                    data = desc[var].flatten().tolist()

                else:
                    data = desc[var]  

                setattr(self.cyInstance, var, data)

            except Exception as e:
                Messages._print(e)
                Messages._print(var, data, type(data))
                Messages._warning('Can not load the variable ' + var + ' in the population ' + self.name)
                Messages._print('Skipping this variable.')
                continue



"""
The CUDA package does contain all code templates required for the code generation in ANNarchy 
targeting NVIDIA graphic cards.

BaseTemplates:

    defines the basic defintions common to all sparse matrix formates, e. g. projection header

[FORMAT]_SingleThread:

    defines the format specific defintions for the currently available formats:

        * COO:          coordinate
        * BSR:          blocked sparse row
        * CSR:          compressed sparse row
        * CSR_Scalar:   a specialized CSR implementation
        * CSR_Vector:   a specialized CSR implementation
        * ELL:          ELLPACK
        * ELLR:         ELLPACK with some GPU-specific optimizations
        * HYB:          a hybrid format using ELLPACK and Coordinate
        * Dense:        a full matrix representation

    there are some special purpose implementations:

        * CSR_T:        csrc (transposed)
"""
from . import COO as COO_CUDA
from . import BSR as BSR_CUDA
from . import CSR as CSR_CUDA
from . import CSR_T as CSR_T_CUDA
from . import CSR_Scalar as CSR_SCALAR_CUDA
from . import CSR_Vector as CSR_VECTOR_CUDA
from . import ELL as ELL_CUDA
from . import ELLR as ELLR_CUDA
from . import SELL as SELL_CUDA
from . import HYB as HYB_CUDA
from . import Dense as Dense_CUDA
from . import Dense_T as Dense_T_CUDA

__all__ = ["BaseTemplates", "COO_CUDA", "BSR_CUDA", "CSR_CUDA", "CSR_T_CUDA", "CSR_SCALAR_CUDA", "CSR_VECTOR_CUDA", "ELL_CUDA", "ELLR_CUDA", "SELL_CUDA", "HYB_CUDA", "Dense_CUDA", "Dense_T_CUDA"]


# Generic imports
import os, sys

# ANNarchy core
from .core.Global import magic_network, clear, check_profile_results
from .core.Constant import Constant
from .core.Neuron import Neuron
from .core.Parameters import Parameter, Variable, Creating, Pruning
from .core.Synapse import Synapse
from .core.Population import Population
from .core.Projection import Projection
from .inputs import *
from .core.Dendrite import Dendrite
from .core.Random import Uniform, DiscreteUniform, Normal, LogNormal, Gamma, Exponential, Binomial
from .core.IO import save, load, load_parameter, load_parameters, save_parameters, MonitorList
from .core.Utils import sparse_random_matrix, sparse_delays_from_weights, timeit
from .core.Monitor import Monitor
from .core.Network import Network
from .parser.report.Report import report
from .models.Neurons import *
from .models.Synapses import *
from .extensions import *

# Cython modules
try:
    # HD: until version 4.6 the connectivity class wasn't named properly. To ensure backward compability
    #     we rename the LILConnectivity to CSR
    from .cython_ext import LILConnectivity
    from .cython_ext import LILConnectivity as CSR
except Exception as e:
    print(e)
    print("""
-------------------------------------------------------------------------------------------------
Warning: Cython modules can not be imported. 
If you are installing ANNarchy, this is normal, ignore this message. 
If ANNarchy is already installed, something went wrong with the compilation, try reinstalling.
-------------------------------------------------------------------------------------------------
""")

# several setup() arguments can be set on command-line
from ANNarchy.generator.CmdLineArgParser import CmdLineArgParser
_arg_parser = CmdLineArgParser()
_arg_parser.parse_arguments_for_setup()

# Automatically call ANNarchy.core.Global.clear()
# if the script terminates
import atexit
atexit.register(check_profile_results)
atexit.register(clear)

# Version
__version__ = '5.0'
__release__ = '5.0.0rc4'

# Bad imports (from ANNarchy 4)
# compile, setup, simulate, simulate_until -> not used anymore
# reset, get_population, get_projection, populations, projections, monitors, enable_learning, disable_learning, get_time, set_time, get_current_step, set_current_step, dt, set_seed -> can now be directly imported from "core.Global"
def __getattr__(name):
    if name == "compile":
        raise ImportError(
            "The function 'compile' cannot be used anymore in ANNarchy 5.\n"
            "Please update your code to use the compile() function of a Network object.\n"
            "In case you defined populations, projections, and monitors as in ANNarchy 4, you can obtain the corresponding network object using the 'magic_network()' function.\n"
        )
    elif name == "setup":
        raise ImportError(
            "The function 'setup' cannot be used anymore in ANNarchy 5.\n"
            "Please update your code to use the config() function of a Network object.\n"
        )
    elif name == "simulate":
        raise ImportError(
            "The function 'simulate' cannot be used anymore in ANNarchy 5.\n"
            "Please update your code to use the simulate() function of a Network object.\n"
        )
    elif name == "simulate_until":
        raise ImportError(
            "The function 'simulate_until' cannot be used anymore in ANNarchy 5.\n"
            "Please update your code to use the simulate_until() function of a Network object.\n"
        )
    elif name in ["reset", "get_population", "get_projection", "populations", "projections", "monitors",
                  "enable_learning", "disable_learning", "get_time", "set_time", "get_current_step",
                  "set_current_step", "dt", "set_seed"]:
        raise ImportError(
            f"The function '{name}' should not be used anymore in ANNarchy 5.\n"
            "Please update your code to use the corresponding function of a Network object.\n"
            "In case you defined a magic network (function 'magic_network') in ANNarchy 4 style, you can still import the function '{name}' from the 'ANNarchy.core.Global' module.\n"
        )
    raise AttributeError(f"module 'ANNarchy' has no attribute '{name}'")

print( 'ANNarchy ' + __version__ + ' (' + __release__ + \
                    ') on ' + sys.platform + ' (' + os.name + ').' )



"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.core.Projection import Projection
from ANNarchy.intern.ConfigManagement import ConfigManager

class SpecificProjection(Projection):
    """
    Interface class for user-defined definition of Projection objects. An inheriting
    class need to override the implementor functions _generate_[paradigm], otherwise
    a NotImplementedError exception will be thrown.
    """
    def __init__(self, pre, post, target, synapse=None, name=None, copied=False, net_id=0):
        """
        Initialization, receive parameters of Projection objects.

        :param pre: pre-synaptic population.
        :param post: post-synaptic population.
        :param target: type of the connection.
        :param window: duration of the time window to collect spikes (default: dt).
        """
        Projection.__init__(self, pre=pre, post=post, target=target, synapse=synapse, name=name, copied=copied, net_id=net_id)

    def _generate(self):
        """
        Overridden method of Population, called during the code generation process.
        This function selects dependent on the chosen paradigm the correct implementor
        functions defined by the user.
        """
        if ConfigManager().get('paradigm', self.net_id) == "openmp":
            if ConfigManager().get('num_threads', self.net_id) == 1:
                self._generate_st()
            else:
                self._generate_omp()
        elif ConfigManager().get('paradigm', self.net_id) == "cuda":
            self._generate_cuda()
        else:
            raise NotImplementedError

    def _generate_omp(self):
        """
        Intended to be overridden by child class. Implememt code adjustments intended for single thread and openMP paradigm.
        """
        raise NotImplementedError

    def _generate_cuda(self):
        """
        Intended to be overridden by child class. Implememt code adjustments intended for CUDA paradigm.
        """
        raise NotImplementedError







"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

from ANNarchy.generator.Utils import tabify
from ANNarchy.intern.ConfigManagement import ConfigManager

from .ProfileGenerator import ProfileGenerator
from .ProfileTemplate import profile_base_template, cpp11_profile_template, cpp11_omp_profile_template, cpp11_profile_header

class CPP11Profile(ProfileGenerator):
    """
    Extent the generated code by profiling annotations using the C++11
    steady clock counter.
    """
    def __init__(self, annarchy_dir, net_id):
        ProfileGenerator.__init__(self, annarchy_dir, net_id)

    def generate(self):
        """
        Generate Profiling class code, called from Generator instance.
        """
        # Generate header for profiling
        with open(self.annarchy_dir+'/generate/net'+str(self._net_id)+'/Profiling.hpp', 'w') as ofile:
            ofile.write(self._generate_header())

    def generate_body_dict(self):
        """
        Creates a dictionary, contain profile code snippets.
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:
            body_dict = {
                'prof_include': cpp11_profile_template['include'],
                'prof_step_pre': cpp11_profile_template['step_pre'],
                'prof_step_post': cpp11_profile_template['step_post'],
                'prof_run_pre': cpp11_profile_template['run_pre'],
                'prof_run_post': cpp11_profile_template['run_post'],
                'prof_proj_psp_pre': cpp11_profile_template['proj_psp_pre'],
                'prof_proj_psp_post': cpp11_profile_template['proj_psp_post'],
                'prof_proj_step_pre': cpp11_profile_template['proj_step_pre'],
                'prof_proj_step_post': cpp11_profile_template['proj_step_post'],
                'prof_proj_post_event_pre': cpp11_profile_template['proj_post_event_pre'],
                'prof_proj_post_event_post': cpp11_profile_template['proj_post_event_post'],
                'prof_neur_step_pre': cpp11_profile_template['neur_step_pre'],
                'prof_neur_step_post': cpp11_profile_template['neur_step_post'],
                'prof_rng_pre': cpp11_profile_template['rng_pre'],
                'prof_rng_post': cpp11_profile_template['rng_post'],
                'prof_record_pre': cpp11_profile_template['record_pre'],
                'prof_record_post': cpp11_profile_template['record_post'],
                'prof_global_ops_pre': cpp11_profile_template['global_op_pre'],
                'prof_global_ops_post': cpp11_profile_template['global_op_post']
            }
        else:
            body_dict = {
                'prof_include': cpp11_omp_profile_template['include'],
                'prof_step_pre': cpp11_omp_profile_template['step_pre'],
                'prof_step_post': cpp11_omp_profile_template['step_post'],
                'prof_run_pre': cpp11_omp_profile_template['run_pre'],
                'prof_run_post': cpp11_omp_profile_template['run_post'],
                'prof_proj_psp_pre': cpp11_omp_profile_template['proj_psp_pre'],
                'prof_proj_psp_post': cpp11_omp_profile_template['proj_psp_post'],
                'prof_proj_post_event_pre': cpp11_omp_profile_template['proj_post_event_pre'],
                'prof_proj_post_event_post': cpp11_omp_profile_template['proj_post_event_post'],
                'prof_proj_step_pre': cpp11_omp_profile_template['proj_step_pre'],
                'prof_proj_step_post': cpp11_omp_profile_template['proj_step_post'],
                'prof_neur_step_pre': cpp11_omp_profile_template['neur_step_pre'],
                'prof_neur_step_post': cpp11_omp_profile_template['neur_step_post'],
                'prof_rng_pre': cpp11_omp_profile_template['rng_pre'],
                'prof_rng_post': cpp11_omp_profile_template['rng_post'],
                'prof_record_pre': cpp11_omp_profile_template['record_pre'],
                'prof_record_post': cpp11_omp_profile_template['record_post'],
                'prof_global_ops_pre': cpp11_omp_profile_template['global_op_pre'],
                'prof_global_ops_post': cpp11_omp_profile_template['global_op_post']
            }

        return body_dict

    def generate_include(self):
        if ConfigManager().get('num_threads', self._net_id) == 1:
            return cpp11_profile_template['include']
        else:
            return cpp11_omp_profile_template['include']

    def generate_init_network(self):
        if ConfigManager().get('num_threads', self._net_id) == 1:
            return cpp11_profile_template['init']
        else:
            return cpp11_omp_profile_template['init']

    def generate_init_population(self, pop):
        """
        Generate initialization code for population
        """
        declare = """
    // Profiling
    Measurement* measure_step;   // update ODE/non-ODE
    Measurement* measure_rng;    // draw random numbers
    Measurement* measure_delay;  // delay variables (in many cases "r")
    Measurement* measure_sc;     // spike condition
"""
        init = """        // Profiling
        measure_step = prof_ptr->register_function("pop", "%(name)s", %(id)s, "step", "%(label)s");
        measure_rng = prof_ptr->register_function("pop", "%(name)s", %(id)s, "rng", "%(label)s");
        measure_delay = prof_ptr->register_function("pop", "%(name)s", %(id)s, "delay", "%(label)s");
        measure_sc = prof_ptr->register_function("pop", "%(name)s", %(id)s, "spike", "%(label)s");
""" % {'name': pop.name, 'id': pop.id, 'label': pop.name}

        return declare, init

    def generate_init_projection(self, proj):
        """
        Generate initialization code for projection
        """
        declare = """
    Measurement* measure_psp;
    Measurement* measure_step;
    Measurement* measure_pe;
"""
        if isinstance(proj.target, str):
            target = proj.target
        else:
            target = proj.target[0]
            for tar in proj.target[1:]:
                target += "_"+tar

        init = """        // Profiling
        measure_psp = prof_ptr->register_function("proj", "%(name)s", %(id_proj)s, "psp", "%(label)s");
        measure_step = prof_ptr->register_function("proj", "%(name)s", %(id_proj)s, "step", "%(label)s");
        measure_pe = prof_ptr->register_function("proj", "%(name)s", %(id_proj)s, "post_event", "%(label)s");
""" % {'id_proj': proj.id, 'name': proj.name, 'label': proj.pre.name+'_'+proj.post.name+'_'+target}

        return declare, init

    def annotate_computesum_rate(self, proj, code):
        """
        annotate the computesum compuation code
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:
            prof_begin = cpp11_profile_template['compute_psp']['before']
            prof_end = cpp11_profile_template['compute_psp']['after']
        else:
            prof_begin = cpp11_omp_profile_template['compute_psp']['before']
            prof_end = cpp11_omp_profile_template['compute_psp']['after']

        prof_code = """
        // first run, measuring average time
%(prof_begin)s
%(code)s
%(prof_end)s
""" % {
        'code': code,
        'prof_begin': tabify(prof_begin, 2),
        'prof_end': tabify(prof_end,2)
        }

        return prof_code

    def annotate_computesum_spiking(self, proj, code):
        """
        annotate the computesum compuation code
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:
            prof_begin = cpp11_profile_template['compute_psp']['before'] % {'name': 'proj'+str(proj.id)}
            prof_end = cpp11_profile_template['compute_psp']['after'] % {'name': 'proj'+str(proj.id)}
        else:
            prof_begin = cpp11_omp_profile_template['compute_psp']['before'] % {'name': 'proj'+str(proj.id)}
            prof_end = cpp11_omp_profile_template['compute_psp']['after'] % {'name': 'proj'+str(proj.id)}

        prof_code = """
        // first run, measuring average time
%(prof_begin)s
%(code)s
%(prof_end)s
""" % {'code': code,
       'prof_begin': tabify(prof_begin, 2),
       'prof_end': tabify(prof_end,2)
       }
        return prof_code

    def annotate_update_synapse(self, proj, code):
        """
        annotate the update synapse code, generated by ProjectionGenerator.update_synapse()
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:        
            prof_begin = cpp11_profile_template['update_synapse']['before']
            prof_end = cpp11_profile_template['update_synapse']['after']
        else:
            prof_begin = cpp11_omp_profile_template['update_synapse']['before']
            prof_end = cpp11_omp_profile_template['update_synapse']['after']

        prof_code = """
// first run, measuring average time
%(prof_begin)s
%(code)s
%(prof_end)s
""" % {'code': code,
       'prof_begin': tabify(prof_begin, 2),
       'prof_end': tabify(prof_end,2)
       }

        return prof_code

    def annotate_post_event(self, proj, code):
        """
        annotate the post-event code
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:
            prof_begin = cpp11_profile_template['post_event']['before']
            prof_end = cpp11_profile_template['post_event']['after']
        else:
            prof_begin = cpp11_omp_profile_template['post_event']['before']
            prof_end = cpp11_omp_profile_template['post_event']['after']

        prof_dict = {
            'code': code,
            'prof_begin': tabify(prof_begin,2),
            'prof_end': tabify(prof_end,2)
        }
        prof_code = """
%(prof_begin)s
%(code)s
%(prof_end)s
""" % prof_dict

        return prof_code

    def annotate_update_neuron(self, pop, code):
        """
        annotate the update neuron code
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:        
            prof_begin = cpp11_profile_template['update_neuron']['before'] % {'name': pop.name}
            prof_end = cpp11_profile_template['update_neuron']['after'] % {'name': pop.name}
        else:
            prof_begin = cpp11_omp_profile_template['update_neuron']['before'] % {'name': pop.name}
            prof_end = cpp11_omp_profile_template['update_neuron']['after'] % {'name': pop.name}

        prof_code = """
        // first run, measuring average time
%(prof_begin)s
%(code)s
%(prof_end)s
""" % {'code': code,
       'prof_begin': tabify(prof_begin, 2),
       'prof_end': tabify(prof_end,2)
       }
        return prof_code

    def annotate_spike_cond(self, pop, code):
        """
        annotate the spike condition code
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:
            prof_begin = cpp11_profile_template['spike_gather']['before'] % {'name': pop.name}
            prof_end = cpp11_profile_template['spike_gather']['after'] % {'name': pop.name}
        else:
            prof_begin = cpp11_omp_profile_template['spike_gather']['before'] % {'name': pop.name}
            prof_end = cpp11_omp_profile_template['spike_gather']['after'] % {'name': pop.name}

        prof_dict = {
            'code': code,
            'prof_begin': tabify(prof_begin,2),
            'prof_end': tabify(prof_end,2)
        }
        prof_code = """
%(prof_begin)s
%(code)s
%(prof_end)s
""" % prof_dict

        return prof_code

    def annotate_update_rng(self, pop, code):
        """
        annotate update rng kernel (only for CPUs available)
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:
            prof_begin = cpp11_profile_template['update_rng']['before'] % {'name': pop.name}
            prof_end = cpp11_profile_template['update_rng']['after'] % {'name': pop.name}
        else:
            prof_begin = cpp11_omp_profile_template['update_rng']['before'] % {'name': pop.name}
            prof_end = cpp11_omp_profile_template['update_rng']['after'] % {'name': pop.name}

        prof_dict = {
            'code': code,
            'prof_begin': tabify(prof_begin,2),
            'prof_end': tabify(prof_end,2)
        }
        prof_code = """
%(prof_begin)s
%(code)s
%(prof_end)s
"""
        return prof_code % prof_dict

    def annotate_update_delay(self, pop, code):
        """
        annotate update delay kernel (only for CPUs available)
        """
        if ConfigManager().get('num_threads', self._net_id) == 1:
            prof_begin = cpp11_profile_template['update_delay']['before'] % {'name': pop.name}
            prof_end = cpp11_profile_template['update_delay']['after'] % {'name': pop.name}
        else:
            prof_begin = cpp11_omp_profile_template['update_delay']['before'] % {'name': pop.name}
            prof_end = cpp11_omp_profile_template['update_delay']['after'] % {'name': pop.name}

        prof_dict = {
            'code': code,
            'prof_begin': tabify(prof_begin,2),
            'prof_end': tabify(prof_end,2)
        }
        prof_code = """
%(prof_begin)s
%(code)s
%(prof_end)s
"""
        return prof_code % prof_dict

    def _generate_header(self):
        """
        generate Profiling.hpp
        """
        config_xml = """
        _out_file << "  <config>" << std::endl;
        _out_file << "    <paradigm>%(paradigm)s</paradigm>" << std::endl;
        _out_file << "    <num_threads>%(num_threads)s</num_threads>" << std::endl;
        _out_file << "  </config>" << std::endl;
        """ % {
            'paradigm': ConfigManager().get('paradigm', self._net_id),
            'num_threads': ConfigManager().get('num_threads', self._net_id)
        }
        config = ConfigManager().get('paradigm', self._net_id) + '_'  + str(ConfigManager().get('num_threads', self._net_id)) + 'threads'

        timer_import = "#include <chrono>"
        timer_start = "std::chrono::time_point<std::chrono::steady_clock> _profiler_start;"
        timer_init = "_profiler_start = std::chrono::steady_clock::now();"
        return profile_base_template % {
            'timer_import': timer_import,
            'timer_start_decl': timer_start,
            'timer_init': timer_init,
            'config': config,
            'result_file': "results_%(config)s.xml" % {'config':config} if ConfigManager().get('profile_out', self._net_id) == None else ConfigManager().get('profile_out', self._net_id),
            'config_xml': config_xml,
            'measurement_class': cpp11_profile_header
        }



from .BoldModel import BoldModel
from .BoldMonitor import BoldMonitor
from .PredefinedModels import balloon_maith2021, balloon_CN, balloon_CL, balloon_RN, balloon_RL, balloon_two_inputs

from .NormProjection import NormProjection

__all__ = ["BoldMonitor", "BoldModel", "balloon_maith2021", "balloon_CN", "balloon_CL", "balloon_RN", "balloon_RL", "balloon_two_inputs", "NormProjection"]


"""
:copyright: Copyright 2013 - now, see AUTHORS.
:license: GPLv2, see LICENSE for details.
"""

basetemplate = """#include <nanobind/nanobind.h>
#include <nanobind/stl/vector.h>
#include <nanobind/stl/string.h>
#include <nanobind/stl/map.h>
#include <nanobind/stl/bind_vector.h>

#include "ANNarchy.hpp"

NB_MODULE(ANNarchyCore%(net_id)s, m) {

    // Global functions
    m.def("set_seed", &setSeed);
    m.def("pyx_create", &create_cpp_instances);
    m.def("pyx_initialize", &initialize);
    m.def("run", &run);
    m.def("run_until", &run_until);
    m.def("step", &step);
    m.def("set_time", &setTime);
    m.def("get_time", &getTime);

    // Target device specific
%(device_specific)s

    // Simulation-related objects
%(functions_wrapper)s

%(constant_wrapper)s

%(pop_struct_wrapper)s

%(proj_struct_wrapper)s

%(pop_mon_wrapper)s

%(proj_mon_wrapper)s

%(profiling_wrapper)s
}
"""
